############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import os

# external packages
import PyQt5.QtWidgets
from astropy.io import fits
from astropy import wcs
from astropy.nddata import Cutout2D
from astropy.visualization import MinMaxInterval
from astropy.visualization import AsinhStretch
from astropy.visualization import imshow_norm
from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt
from skyfield.api import Angle
import numpy as np
from colour_demosaicing import demosaicing_CFA_Bayer_bilinear
from mountcontrol.convert import convertToDMS, convertToHMS
import sep

# local import
from gui.utilities import toolsQtWidget
from gui.widgets import image_ui
from base.tpool import Worker


class ImageWindowSignals(PyQt5.QtCore.QObject):
    """
    """
    __all__ = ['ImageWindowSignals']
    showCurrent = PyQt5.QtCore.pyqtSignal()
    solveImage = PyQt5.QtCore.pyqtSignal(object)


class ImageWindow(toolsQtWidget.MWidget):
    """
    """
    __all__ = ['ImageWindow']

    def __init__(self, app):
        super().__init__()

        self.app = app
        self.threadPool = app.threadPool
        self.ui = image_ui.Ui_ImageDialog()
        self.ui.setupUi(self)
        self.initUI()
        self.signals = ImageWindowSignals()

        self.imageFileName = ''
        self.imageFileNameOld = ''
        self.expTime = 1
        self.binning = 1
        self.image = None
        self.folder = ''
        self.header = None
        self.imageStack = None
        self.numberStack = 0
        self.colorMap = None
        self.stretch = None
        self.objs = None
        self.bk_back = None
        self.bk_rms = None
        self.flux = None
        self.radius = None

        self.deviceStat = {
            'expose': False,
            'exposeN': False,
            'solve': False,
        }
        self.view = {0: 'Image Raw',
                     1: 'Image with Sources',
                     2: 'Photometry: HFD value',
                     3: 'Photometry: Eccentricity',
                     4: 'Photometry: Background level',
                     5: 'Photometry: Background noise',
                     6: 'Photometry: Flux',
                     }

        self.colorMaps = {'Grey': 'gray',
                          'Cool': 'plasma',
                          'Rainbow': 'rainbow',
                          'Spectral': 'nipy_spectral',
                          }

        self.stretchValues = {'Low XX': 0.2,
                              'Low X': 0.1,
                              'Low': 0.05,
                              'Mid': 0.025,
                              'High': 0.005,
                              'Super ': 0.0025,
                              'Super X': 0.001,
                              'Super XX': 0.0005,
                              }

        self.zoomLevel = {' 1x Zoom': 1,
                          ' 2x Zoom': 2,
                          ' 4x Zoom': 4,
                          ' 8x Zoom': 8,
                          '16x Zoom': 16,
                          }

        self.imageMat = self.embedMatplot(self.ui.image, constrainedLayout=False)
        self.imageMat.parentWidget().setStyleSheet(self.BACK_BG)
        self.fig = self.imageMat.figure
        self.axe = None
        self.axeCB = None

        self.app.update1s.connect(self.updateWindowsStats)

    def initConfig(self):
        """
        :return: True for test purpose
        """
        if 'imageW' not in self.app.config:
            self.app.config['imageW'] = {}
        config = self.app.config['imageW']
        x = config.get('winPosX', 80)
        y = config.get('winPosY', 80)
        if x > self.screenSizeX:
            x = 0
        if y > self.screenSizeY:
            y = 0
        self.move(x, y)
        height = config.get('height', 600)
        width = config.get('width', 800)
        self.resize(width, height)

        self.setupDropDownGui()
        self.ui.color.setCurrentIndex(config.get('color', 0))
        self.ui.zoom.setCurrentIndex(config.get('zoom', 0))
        self.ui.view.setCurrentIndex(config.get('view', 0))

        self.ui.stretch.setCurrentIndex(config.get('stretch', 0))
        self.imageFileName = config.get('imageFileName', '')
        self.folder = self.app.mwGlob.get('imageDir', '')
        self.ui.checkStackImages.setChecked(config.get('checkStackImages', False))
        self.ui.checkShowCrosshair.setChecked(config.get('checkShowCrosshair', False))
        self.ui.checkShowGrid.setChecked(config.get('checkShowGrid', True))
        self.ui.checkAutoSolve.setChecked(config.get('checkAutoSolve', False))
        self.ui.checkEmbedData.setChecked(config.get('checkEmbedData', False))
        return True

    def storeConfig(self):
        """
        :return: True for test purpose
        """
        if 'imageW' not in self.app.config:
            self.app.config['imageW'] = {}
        config = self.app.config['imageW']
        config['winPosX'] = self.pos().x()
        config['winPosY'] = self.pos().y()
        config['height'] = self.height()
        config['width'] = self.width()
        config['color'] = self.ui.color.currentIndex()
        config['zoom'] = self.ui.zoom.currentIndex()
        config['view'] = self.ui.view.currentIndex()
        config['stretch'] = self.ui.stretch.currentIndex()
        config['imageFileName'] = self.imageFileName
        config['checkStackImages'] = self.ui.checkStackImages.isChecked()
        config['checkShowCrosshair'] = self.ui.checkShowCrosshair.isChecked()
        config['checkShowGrid'] = self.ui.checkShowGrid.isChecked()
        config['checkAutoSolve'] = self.ui.checkAutoSolve.isChecked()
        config['checkEmbedData'] = self.ui.checkEmbedData.isChecked()
        return True

    def showWindow(self):
        """
        showWindow prepares all data for showing the window, plots the image and
        show it. afterwards all necessary signal / slot connections will be
        established.

        :return: true for test purpose
        """
        self.show()
        self.showCurrent()

        self.ui.load.clicked.connect(self.selectImage)
        self.ui.color.currentIndexChanged.connect(self.preparePlot)
        self.ui.stretch.currentIndexChanged.connect(self.preparePlot)
        self.ui.zoom.currentIndexChanged.connect(self.showCurrent)
        self.ui.view.currentIndexChanged.connect(self.preparePlot)
        self.ui.checkUseWCS.clicked.connect(self.preparePlot)
        self.ui.checkShowGrid.clicked.connect(self.preparePlot)
        self.ui.checkShowCrosshair.clicked.connect(self.preparePlot)
        self.ui.solve.clicked.connect(self.solveCurrent)
        self.ui.expose.clicked.connect(self.exposeImage)
        self.ui.exposeN.clicked.connect(self.exposeImageN)
        self.ui.abortImage.clicked.connect(self.abortImage)
        self.ui.abortSolve.clicked.connect(self.abortSolve)
        self.signals.showCurrent.connect(self.showCurrent)
        self.signals.solveImage.connect(self.solveImage)
        self.app.showImage.connect(self.showImage)
        return True

    def closeEvent(self, closeEvent):
        """
        closeEvent overlays the window close event of qt. first it stores all
        persistent data for the windows and its functions, than removes al signal
        / slot connections removes the matplotlib embedding and finally calls the
        parent calls for handling the framework close event.

        :param closeEvent:
        :return: True for test purpose
        """
        self.storeConfig()

        self.ui.load.clicked.disconnect(self.selectImage)
        self.ui.color.currentIndexChanged.disconnect(self.preparePlot)
        self.ui.stretch.currentIndexChanged.disconnect(self.preparePlot)
        self.ui.zoom.currentIndexChanged.disconnect(self.showCurrent)
        self.ui.view.currentIndexChanged.disconnect(self.preparePlot)
        self.ui.checkUseWCS.clicked.disconnect(self.preparePlot)
        self.ui.checkShowGrid.clicked.disconnect(self.preparePlot)
        self.ui.checkShowCrosshair.clicked.disconnect(self.preparePlot)
        self.ui.solve.clicked.disconnect(self.solveCurrent)
        self.ui.expose.clicked.disconnect(self.exposeImage)
        self.ui.exposeN.clicked.disconnect(self.exposeImageN)
        self.ui.abortImage.clicked.disconnect(self.abortImage)
        self.ui.abortSolve.clicked.disconnect(self.abortSolve)
        self.signals.showCurrent.disconnect(self.showCurrent)
        self.signals.solveImage.disconnect(self.solveImage)

        self.app.showImage.disconnect(self.showImage)

        plt.close(self.imageMat.figure)
        super().closeEvent(closeEvent)

    def setupDropDownGui(self):
        """
        :return: success for test
        """
        self.ui.color.clear()
        self.ui.color.setView(PyQt5.QtWidgets.QListView())
        for text in self.colorMaps:
            self.ui.color.addItem(text)

        self.ui.zoom.clear()
        self.ui.zoom.setView(PyQt5.QtWidgets.QListView())
        for text in self.zoomLevel:
            self.ui.zoom.addItem(text)

        self.ui.stretch.clear()
        self.ui.stretch.setView(PyQt5.QtWidgets.QListView())
        for text in self.stretchValues:
            self.ui.stretch.addItem(text)

        self.ui.view.clear()
        self.ui.view.setView(PyQt5.QtWidgets.QListView())
        for menuNumber in self.view:
            self.ui.view.addItem(self.view[menuNumber])

        return True

    def updateWindowsStats(self):
        """
        :return: true for test purpose
        """
        if self.deviceStat['expose']:
            self.ui.exposeN.setEnabled(False)
            self.ui.load.setEnabled(False)
            self.ui.abortImage.setEnabled(True)

        elif self.deviceStat['exposeN']:
            self.ui.expose.setEnabled(False)
            self.ui.load.setEnabled(False)
            self.ui.abortImage.setEnabled(True)

        else:
            self.ui.solve.setEnabled(True)
            self.ui.expose.setEnabled(True)
            self.ui.exposeN.setEnabled(True)
            self.ui.load.setEnabled(True)
            self.ui.abortImage.setEnabled(False)

        if self.deviceStat['solve']:
            self.ui.abortSolve.setEnabled(True)

        else:
            self.ui.abortSolve.setEnabled(False)

        if not self.app.deviceStat['camera']:
            self.ui.expose.setEnabled(False)
            self.ui.exposeN.setEnabled(False)

        if not self.app.deviceStat['astrometry']:
            self.ui.solve.setEnabled(False)

        if self.deviceStat['expose']:
            self.changeStyleDynamic(self.ui.expose, 'running', 'true')

        elif self.deviceStat['exposeN']:
            self.changeStyleDynamic(self.ui.exposeN, 'running', 'true')

        else:
            self.changeStyleDynamic(self.ui.expose, 'running', 'false')
            self.changeStyleDynamic(self.ui.exposeN, 'running', 'false')

        if self.deviceStat['solve']:
            self.changeStyleDynamic(self.ui.solve, 'running', 'true')

        else:
            self.changeStyleDynamic(self.ui.solve, 'running', 'false')
        return True

    def selectImage(self):
        """
        :return: success
        """
        val = self.openFile(self, 'Select image file', self.folder,
                            'FITS files (*.fit*)', enableDir=True)
        loadFilePath, name, ext = val
        if not name:
            return False

        self.ui.imageFileName.setText(name)
        self.imageFileName = loadFilePath
        self.app.message.emit(f'Image [{name}] selected', 0)
        self.folder = os.path.dirname(loadFilePath)
        self.app.showImage.emit(self.imageFileName)
        return True

    def setupDistorted(self):
        """
        setupDistorted tries to setup all necessary context for displaying the
        image with wcs distorted coordinates.

        :return: true for test purpose
        """
        self.fig.clf()
        self.axe = self.fig.add_subplot(1, 1, 1,
                                        projection=wcs.WCS(self.header, relax=True),
                                        facecolor=self.M_GREY_LIGHT)

        self.fig.subplots_adjust(bottom=0.1, top=0.9, left=0.1, right=0.85)
        self.axeCB = None
        self.axe.coords.frame.set_color(self.M_BLUE)

        axe0 = self.axe.coords[0]
        axe1 = self.axe.coords[1]

        if self.ui.checkShowGrid.isChecked():
            axe0.grid(True, color=self.M_BLUE, ls='solid', alpha=0.5)
            axe1.grid(True, color=self.M_BLUE, ls='solid', alpha=0.5)

        axe0.tick_params(colors=self.M_BLUE, labelsize=12)
        axe1.tick_params(colors=self.M_BLUE, labelsize=12)
        axe0.set_axislabel('Right Ascension',
                           color=self.M_BLUE,
                           fontsize=12,
                           fontweight='bold')
        axe1.set_axislabel('Declination',
                           color=self.M_BLUE,
                           fontsize=12,
                           fontweight='bold')
        axe0.set_ticks(number=10)
        axe1.set_ticks(number=10)
        axe0.set_ticks_position('lr')
        axe0.set_ticklabel_position('lr')
        axe0.set_axislabel_position('lr')
        axe1.set_ticks_position('tb')
        axe1.set_ticklabel_position('tb')
        axe1.set_axislabel_position('tb')
        return True

    def setupNormal(self):
        """
        :return: True for test purpose
        """
        self.fig.clf()
        self.axe = self.fig.add_subplot(1, 1, 1,
                                        facecolor=self.M_GREY_LIGHT)

        self.fig.subplots_adjust(bottom=0.1, top=0.9, left=0.1, right=0.85)
        self.axeCB = self.fig.add_axes([0.88, 0.1, 0.02, 0.8])
        self.axe.spines['bottom'].set_color(self.M_BLUE)
        self.axe.spines['top'].set_color(self.M_BLUE)
        self.axe.spines['left'].set_color(self.M_BLUE)
        self.axe.spines['right'].set_color(self.M_BLUE)

        factor = self.zoomLevel[self.ui.zoom.currentText()]
        sizeX = self.header.get('NAXIS1', 1) / factor
        sizeY = self.header.get('NAXIS2', 1) / factor
        midX = int(sizeX / 2)
        midY = int(sizeY / 2)
        number = 10

        if self.ui.checkShowCrosshair.isChecked():
            self.axe.axvline(midX, color=self.M_RED)
            self.axe.axhline(midY, color=self.M_RED)

        if self.ui.checkShowGrid.isChecked():
            self.axe.grid(True, color=self.M_BLUE, ls='solid', alpha=0.5)

        self.axe.tick_params(axis='x', which='major',
                             colors=self.M_BLUE, labelsize=12)
        self.axe.tick_params(axis='y', which='major',
                             colors=self.M_BLUE, labelsize=12)

        valueX, _ = np.linspace(-midX, midX, num=number, retstep=True)
        textX = list((str(int(x)) for x in valueX))
        ticksX = list((x + midX for x in valueX))
        self.axe.set_xticks(ticksX)
        self.axe.set_xticklabels(textX)

        valueY, _ = np.linspace(-midY, midY, num=number, retstep=True)
        textY = list((str(int(x)) for x in valueY))
        ticksY = list((x + midY for x in valueY))
        self.axe.set_yticks(ticksY)
        self.axe.set_yticklabels(textY)

        self.axe.set_xlabel(xlabel='Pixel', color=self.M_BLUE,
                            fontsize=12, fontweight='bold')
        self.axe.set_ylabel(ylabel='Pixel', color=self.M_BLUE,
                            fontsize=12, fontweight='bold')

        self.axe.set_xlim(0, sizeX)
        self.axe.set_ylim(0, sizeY)
        return True

    def colorImage(self):
        """
        :return: True
        """
        fallback = list(self.colorMaps.keys())[0]
        self.colorMap = self.colorMaps.get(self.ui.color.currentText(), fallback)
        return True

    def stretchImage(self):
        """
        :return: True
        """
        fallback = list(self.stretchValues.keys())[0]
        value = self.stretchValues.get(self.ui.stretch.currentText(),
                                       self.stretchValues[fallback])
        self.stretch = AsinhStretch(a=value)
        return True

    def imagePlot(self):
        """
        :return:
        """
        if self.objs is not None:
            imageDisp = self.image - self.bk_back

        else:
            imageDisp = self.image

        if self.ui.view.currentIndex() in [0, 1, 2, 3]:
            imageDisp[imageDisp < 0] = 0
            img = imshow_norm(imageDisp,
                              ax=self.axe,
                              origin='lower',
                              interval=MinMaxInterval(),
                              stretch=self.stretch,
                              cmap=self.colorMap,
                              aspect='auto')

            if self.axeCB:
                colorbar = self.fig.colorbar(img[0], cax=self.axeCB)
                colorbar.set_label('Value [pixel value]',
                                   color=self.M_BLUE,
                                   fontsize=12)
                yTicks = plt.getp(colorbar.ax.axes, 'yticklabels')
                plt.setp(yTicks, color=self.M_BLUE, fontweight='bold')

        if self.ui.view.currentIndex() == 1 and self.objs is not None:
            for i in range(len(self.objs)):
                e = Ellipse(xy=(self.objs['x'][i], self.objs['y'][i]),
                            width=6 * self.objs['a'][i],
                            height=6 * self.objs['b'][i],
                            angle=self.objs['theta'][i] * 180. / np.pi)
                e.set_facecolor('none')
                e.set_edgecolor(self.M_BLUE)
                self.axe.add_artist(e)

        if self.ui.view.currentIndex() == 2 and self.radius is not None:
            draw = self.radius.argsort()[-50:][::-1]
            for i in draw:
                e = Ellipse(xy=(self.objs['x'][i], self.objs['y'][i]),
                            width=6 * self.objs['a'][i],
                            height=6 * self.objs['b'][i],
                            angle=self.objs['theta'][i] * 180.0 / np.pi)
                e.set_facecolor('none')
                e.set_edgecolor(self.M_BLUE)
                self.axe.add_artist(e)
                posX = self.objs['x'][i] + self.objs['a'][i] + 5
                posY = self.objs['b'][i] + self.objs['y'][i] + 5
                self.axe.annotate(f'{self.radius[i]:2.1f}',
                                  xy=(posX, posY),
                                  color=self.M_BLUE,
                                  fontweight='bold')

        if self.ui.view.currentIndex() == 3 and self.objs is not None:
            a = self.objs['a']
            b = self.objs['b']
            eccentricity = np.sqrt(1 - b ** 2 / a ** 2)
            area = 5
            scatter = self.axe.scatter(self.objs['x'], self.objs['y'],
                                       c=eccentricity, s=area, cmap='gnuplot2')

            colorbar = self.fig.colorbar(scatter, cax=self.axeCB)
            colorbar.set_label('Eccentricity []', color=self.M_BLUE,
                               fontsize=12)
            yTicks = plt.getp(colorbar.ax.axes, 'yticklabels')
            plt.setp(yTicks, color=self.M_BLUE, fontweight='bold')

        if self.ui.view.currentIndex() == 4 and self.bk_back is not None:
            img = imshow_norm(self.bk_back,
                              ax=self.axe,
                              origin='lower',
                              interval=MinMaxInterval(),
                              stretch=self.stretch,
                              cmap=self.colorMap,
                              aspect='auto')
            if self.axeCB:
                colorbar = self.fig.colorbar(img[0], cax=self.axeCB)
                colorbar.set_label('Background level [adu]',
                                   color=self.M_BLUE,
                                   fontsize=12)
                yTicks = plt.getp(colorbar.ax.axes, 'yticklabels')
                plt.setp(yTicks, color=self.M_BLUE, fontweight='bold')

        if self.ui.view.currentIndex() == 5 and self.bk_rms is not None:
            img = imshow_norm(self.bk_rms,
                              ax=self.axe,
                              origin='lower',
                              interval=MinMaxInterval(),
                              stretch=self.stretch,
                              cmap=self.colorMap,
                              aspect='auto')
            if self.axeCB:
                colorbar = self.fig.colorbar(img[0], cax=self.axeCB)
                colorbar.set_label('Background noise [adu]',
                                   color=self.M_BLUE,
                                   fontsize=12)
                yTicks = plt.getp(colorbar.ax.axes, 'yticklabels')
                plt.setp(yTicks, color=self.M_BLUE,
                         fontweight='bold')

        if self.ui.view.currentIndex() == 6 and self.flux is not None:
            flux = np.log(self.flux)
            area = 3 * flux
            scatter = self.axe.scatter(self.objs['x'], self.objs['y'],
                                       c=flux, s=area, cmap=self.colorMap)

            colorbar = self.fig.colorbar(scatter, cax=self.axeCB)
            colorbar.set_label('Flux [log(x)]', color=self.M_BLUE, fontsize=12)
            yTicks = plt.getp(colorbar.ax.axes, 'yticklabels')
            plt.setp(yTicks, color=self.M_BLUE, fontweight='bold')

        self.axe.figure.canvas.draw()

        return True

    def writeHeaderDataToGUI(self):
        """
        :return: True for test purpose
        """
        name = self.header.get('OBJECT', '').upper()
        self.ui.object.setText(f'{name}')

        ra = Angle(degrees=self.header.get('RA', 0))
        dec = Angle(degrees=self.header.get('DEC', 0))

        # ra will be in hours
        self.ui.ra.setText(f'{ra.hstr(warn=False)}')
        self.ui.dec.setText(f'{dec.dstr()}')

        scale = self.header.get('SCALE', 0)
        rotation = self.header.get('ANGLE', 0)
        self.ui.scale.setText(f'{scale:5.3f}')
        self.ui.rotation.setText(f'{rotation:6.3f}')

        ccdTemp = self.header.get('CCD-TEMP', 0)
        self.ui.ccdTemp.setText(f'{ccdTemp:4.1f}')

        expTime1 = self.header.get('EXPOSURE', 0)
        expTime2 = self.header.get('EXPTIME', 0)
        expTime = max(expTime1, expTime2)
        self.ui.expTime.setText(f'{expTime:5.1f}')

        filterCCD = self.header.get('FILTER', 0)
        self.ui.filter.setText(f'{filterCCD}')

        binX = self.header.get('XBINNING', 0)
        binY = self.header.get('YBINNING', 0)
        self.ui.binX.setText(f'{binX:1.0f}')
        self.ui.binY.setText(f'{binY:1.0f}')

        sqm = max(self.header.get('SQM', 0),
                  self.header.get('SKY-QLTY', 0),
                  self.header.get('MPSAS', 0),
                  )
        self.ui.sqm.setText(f'{sqm:5.2f}')

        flipped = bool(self.header.get('FLIPPED', False))
        self.ui.isFlipped.setEnabled(flipped)

        return True

    def preparePlot(self):
        """
        :return:
        """
        if self.image is None:
            return False

        if self.header is None:
            return False

        self.updateWindowsStats()
        if 'CTYPE1' in self.header:
            wcsObject = wcs.WCS(self.header, relax=True)
            hasCelestial = wcsObject.has_celestial
            hasDistortion = wcsObject.has_distortion

        else:
            hasCelestial = False
            hasDistortion = False

        self.ui.hasDistortion.setEnabled(hasDistortion)
        self.ui.checkUseWCS.setEnabled(hasDistortion)
        self.ui.hasWCS.setEnabled(hasCelestial)

        canWCS = self.ui.view.currentIndex() in [0, 1, 2]
        useWCS = self.ui.checkUseWCS.isChecked()

        if hasDistortion and useWCS and canWCS:
            self.setupDistorted()

        else:
            self.setupNormal()

        self.writeHeaderDataToGUI()
        self.stretchImage()
        self.colorImage()
        self.imagePlot()
        return True

    def workerPhotometry(self):
        """
        :return:
        """
        bkg = sep.Background(self.image)
        self.bk_back = bkg.back()
        self.bk_rms = bkg.rms()
        image_sub = self.image - bkg
        self.objs = sep.extract(image_sub, 1.5, err=bkg.globalrms)
        self.flux, _, _ = sep.sum_circle(image_sub,
                                         self.objs['x'],
                                         self.objs['y'],
                                         3.0,
                                         err=bkg.globalrms,
                                         gain=1.0)
        self.radius, _ = sep.flux_radius(image_sub,
                                         self.objs['x'],
                                         self.objs['y'],
                                         6.0 * self.objs['a'],
                                         0.5,
                                         normflux=self.flux,
                                         subpix=5)
        return True

    def prepareImageForPhotometry(self):
        """
        :return:
        """
        if not self.objs:
            worker = Worker(self.workerPhotometry)
            worker.signals.finished.connect(self.preparePlot)
            self.threadPool.start(worker)

        self.preparePlot()
        return True

    def stackImages(self):
        """
        :return:
        """
        if not self.ui.checkStackImages.isChecked():
            self.imageStack = None
            self.ui.numberStacks.setText('single')
            return False

        if np.shape(self.image) != np.shape(self.imageStack):
            self.imageStack = None

        self.ui.numberStacks.setText(f'mean of: {self.numberStack:4.0f}')

        if self.imageStack is None:
            self.imageStack = self.image
            self.numberStack = 1

        else:
            self.imageStack = np.add(self.imageStack, self.image)
            self.numberStack += 1

        self.image = self.imageStack / self.numberStack
        return True

    def zoomImage(self):
        """
        :return: true
        """
        sizeY, sizeX = self.image.shape
        fallback = list(self.zoomLevel.keys())[0]
        factor = self.zoomLevel.get(self.ui.zoom.currentText(), fallback)
        if factor == 1:
            return

        position = (int(sizeX / 2), int(sizeY / 2))
        size = (int(sizeY / factor), int(sizeX / factor))
        self.image = Cutout2D(self.image,
                              position=position,
                              size=size,
                              wcs=wcs.WCS(self.header, relax=True),
                              copy=True).data
        return True

    def showImage(self, imagePath=''):
        """
        tho idea of processing the image data until it is shown on screen is
        to split the pipeline in several atomic steps, which follow each other.
        each step is calling the next one and depending which user interaction is
        done and how the use case for changing the view is the entrance of the
        pipeline is chosen adequately.

        so we have step 1 loading data and processing if to final monochrome and
        sized format
        - loading data
        - debayer if necessary
        - getting header data and cleaning it up
        - zoom and stack

        :param imagePath:
        :return:
        """
        if not imagePath:
            return False

        if not os.path.isfile(imagePath):
            return False

        self.imageFileName = imagePath
        full, short, ext = self.extractNames([imagePath])
        self.ui.imageFileName.setText(short)

        with fits.open(imagePath, mode='update') as fitsHandle:
            self.image = fitsHandle[0].data
            self.header = fitsHandle[0].header

        if self.image is None:
            return False

        if self.header is None:
            return False

        self.image = self.image.astype('float32')

        if 'BAYERPAT' in self.header:
            self.image = demosaicing_CFA_Bayer_bilinear(self.image)
            self.image = np.dot(self.image, [0.2989, 0.5870, 0.1140])

        # correct faulty headers, because some imaging programs did not
        # interpret the Keywords in the right manner (SGPro)
        if self.header.get('CTYPE1', '').endswith('DEF'):
            self.header['CTYPE1'] = self.header['CTYPE1'].replace('DEF', 'TAN')

        if self.header.get('CTYPE2', '').endswith('DEF'):
            self.header['CTYPE2'] = self.header['CTYPE2'].replace('DEF', 'TAN')

        self.zoomImage()
        self.stackImages()
        self.objs = None
        self.prepareImageForPhotometry()
        return True

    def showCurrent(self):
        """
        :return: true for test purpose
        """
        self.showImage(self.imageFileName)
        return True

    def exposeRaw(self):
        """
        :return: True for test purpose
        """
        subFrame = self.app.camera.subFrame
        fastReadout = self.app.camera.checkFastDownload

        time = self.app.mount.obsSite.timeJD.utc_strftime('%Y-%m-%d-%H-%M-%S')
        fileName = time + '-exposure.fits'
        imagePath = self.app.mwGlob['imageDir'] + '/' + fileName
        focalLength = self.app.telescope.focalLength

        self.imageFileNameOld = self.imageFileName

        self.app.camera.expose(imagePath=imagePath,
                               expTime=self.expTime,
                               binning=self.binning,
                               subFrame=subFrame,
                               fastReadout=fastReadout,
                               focalLength=focalLength
                               )

        text = f'Exposing:            [{os.path.basename(imagePath)}]'
        self.app.message.emit(text, 0)
        text = f'Duration:{self.expTime:3.0f}s  '
        text += f'Bin:{self.binning:1.0f}  Sub:{subFrame:3.0f}%'
        self.app.message.emit(f'                     {text}', 0)

        return True

    def exposeImageDone(self, imagePath=''):
        """
        :param imagePath:
        :return: True for test purpose
        """

        self.deviceStat['expose'] = False
        self.app.camera.signals.saved.disconnect(self.exposeImageDone)
        text = f'Exposed:             [{os.path.basename(imagePath)}]'
        self.app.message.emit(text, 0)

        if self.ui.checkAutoSolve.isChecked():
            self.signals.solveImage.emit(imagePath)

        else:
            self.app.showImage.emit(imagePath)

        return True

    def exposeImage(self):
        """
        :return: success
        """
        self.expTime = self.app.camera.expTime
        self.binning = self.app.camera.binning
        self.imageStack = None
        self.deviceStat['expose'] = True
        self.ui.checkStackImages.setChecked(False)
        self.app.camera.signals.saved.connect(self.exposeImageDone)
        self.exposeRaw()
        return True

    def exposeImageNDone(self, imagePath=''):
        """
        :param imagePath:
        :return: True for test purpose
        """
        text = f'Exposed:            [{os.path.basename(imagePath)}]'
        self.app.message.emit(text, 0)

        if self.ui.checkAutoSolve.isChecked():
            self.signals.solveImage.emit(imagePath)

        else:
            self.app.showImage.emit(imagePath)

        self.exposeRaw()
        return True

    def exposeImageN(self):
        """
        :return: success
        """
        self.expTime = self.app.camera.expTimeN
        self.binning = self.app.camera.binningN
        self.imageStack = None
        self.deviceStat['exposeN'] = True
        self.app.camera.signals.saved.connect(self.exposeImageNDone)
        self.exposeRaw()
        return True

    def abortImage(self):
        """
        :return: True for test purpose
        """
        self.app.camera.abort()
        # for disconnection we have to split which slots were connected to
        # disable the right ones
        if self.deviceStat['expose']:
            self.app.camera.signals.saved.disconnect(self.exposeImageDone)

        if self.deviceStat['exposeN']:
            self.app.camera.signals.saved.disconnect(self.exposeImageNDone)

        # last image file was not stored, so getting last valid it back
        self.imageFileName = self.imageFileNameOld
        self.deviceStat['expose'] = False
        self.deviceStat['exposeN'] = False
        self.app.message.emit('Exposing aborted', 2)
        return True

    def solveDone(self, result=None):
        """
        :param result: result (named tuple)
        :return: success
        """
        self.deviceStat['solve'] = False
        self.app.astrometry.signals.done.disconnect(self.solveDone)

        if not result:
            self.app.message.emit('Solving error, result missing', 2)
            return False

        if result['success']:
            text = 'Solved :             '
            text += f'RA: {convertToHMS(result["raJ2000S"])} '
            text += f'({result["raJ2000S"].hours:4.3f}), '
            text += f'DEC: {convertToDMS(result["decJ2000S"])} '
            text += f'({result["decJ2000S"].degrees:4.3f}), '
            self.app.message.emit(text, 0)
            text = '                     '
            text += f'Angle: {result["angleS"]:3.0f}, '
            text += f'Scale: {result["scaleS"]:4.3f}, '
            text += f'Error: {result["errorRMS_S"]:4.1f}'
            self.app.message.emit(text, 0)

        else:
            text = f'Solving error:       {result.get("message")}'
            self.app.message.emit(text, 2)
            return False

        isStack = self.ui.checkStackImages.isChecked()
        isAutoSolve = self.ui.checkAutoSolve.isChecked()

        if not isStack or isAutoSolve:
            self.app.showImage.emit(result['solvedPath'])
        return True

    def solveImage(self, imagePath=''):
        """
        :param imagePath:
        :return:
        """
        if not imagePath:
            return False

        if not os.path.isfile(imagePath):
            return False

        updateFits = self.ui.checkEmbedData.isChecked()
        self.app.astrometry.solveThreading(fitsPath=imagePath,
                                           updateFits=updateFits,
                                           )
        self.deviceStat['solve'] = True
        self.app.astrometry.signals.done.connect(self.solveDone)
        text = f'Solving:             [{os.path.basename(imagePath)}]'
        self.app.message.emit(text, 0)
        return True

    def solveCurrent(self):
        """
        :return: true for test purpose
        """
        self.signals.solveImage.emit(self.imageFileName)
        return True

    def abortSolve(self):
        """
        :return: success
        """
        suc = self.app.astrometry.abort()
        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
import base.packageConfig as pConf
# standard libraries
import time
import subprocess
import sys

# external packages
from pkg_resources import working_set
from distutils.version import StrictVersion
import PyQt5
if pConf.isAvailable:
    import PyQt5.QtMultimedia
import requests
import importlib_metadata
from astropy.utils import iers

# local import
from base.loggerMW import setCustomLoggingLevel
from base import tpool


class SettMisc(object):
    """
    """

    def __init__(self):
        self.audioSignalsSet = dict()
        self.guiAudioList = dict()
        self.process = None
        self.mutexInstall = PyQt5.QtCore.QMutex()

        self.app.mount.signals.alert.connect(lambda: self.playSound('MountAlert'))
        self.app.dome.signals.slewFinished.connect(lambda: self.playSound('DomeSlew'))
        self.app.mount.signals.slewFinished.connect(lambda: self.playSound('MountSlew'))
        self.app.camera.signals.saved.connect(lambda: self.playSound('ImageSaved'))
        self.app.astrometry.signals.done.connect(lambda: self.playSound('ImageSolved'))

        self.ui.loglevelDebugTrace.clicked.connect(self.setLoggingLevel)
        self.ui.loglevelDebug.clicked.connect(self.setLoggingLevel)
        self.ui.loglevelStandard.clicked.connect(self.setLoggingLevel)
        self.ui.isOnline.clicked.connect(self.setWeatherOnline)
        self.ui.isOnline.clicked.connect(self.setupIERS)
        self.ui.versionBeta.clicked.connect(self.showUpdates)
        self.ui.versionRelease.clicked.connect(self.showUpdates)
        self.ui.isOnline.clicked.connect(self.showUpdates)
        self.ui.installVersion.clicked.connect(self.installVersion)

        self.setupAudioSignals()

    def initConfig(self):
        """
        :return: True for test purpose
        """
        config = self.app.config['mainW']
        self.setupAudioGui()
        self.ui.loglevelDebugTrace.setChecked(config.get('loglevelDebugTrace', False))
        self.ui.loglevelDebug.setChecked(config.get('loglevelDebug', False))
        self.ui.loglevelStandard.setChecked(config.get('loglevelStandard', True))
        self.ui.isOnline.setChecked(config.get('isOnline', False))
        self.ui.soundMountSlewFinished.setCurrentIndex(config.get('soundMountSlewFinished', 0))
        self.ui.soundDomeSlewFinished.setCurrentIndex(config.get('soundDomeSlewFinished', 0))
        self.ui.soundMountAlert.setCurrentIndex(config.get('soundMountAlert', 0))
        self.ui.soundModelingFinished.setCurrentIndex(config.get('soundModelingFinished', 0))
        self.ui.soundImageSaved.setCurrentIndex(config.get('soundImageSaved', 0))
        self.ui.soundImageSolved.setCurrentIndex(config.get('soundImageSolved', 0))

        self.setLoggingLevel()
        self.setWeatherOnline()
        self.setupIERS()

        self.showUpdates()

        return True

    def storeConfig(self):
        """
        :return: True for test purpose
        """
        config = self.app.config['mainW']
        config['loglevelDebugTrace'] = self.ui.loglevelDebugTrace.isChecked()
        config['loglevelDebug'] = self.ui.loglevelDebug.isChecked()
        config['loglevelStandard'] = self.ui.loglevelStandard.isChecked()
        config['isOnline'] = self.ui.isOnline.isChecked()
        config['soundMountSlewFinished'] = self.ui.soundMountSlewFinished.currentIndex()
        config['soundDomeSlewFinished'] = self.ui.soundDomeSlewFinished.currentIndex()
        config['soundMountAlert'] = self.ui.soundMountAlert.currentIndex()
        config['soundModelingFinished'] = self.ui.soundModelingFinished.currentIndex()
        config['soundImageSaved'] = self.ui.soundImageSaved.currentIndex()
        config['soundImageSolved'] = self.ui.soundImageSolved.currentIndex()

        return True

    def setWeatherOnline(self):
        """
        :return: success
        """
        weather = self.app.onlineWeather
        if not weather:
            return False

        weather.online = self.ui.isOnline.isChecked()
        return True

    def setupIERS(self):
        """
        setupIERS enables or disables the update of astropy necessary files
        depending on online status .

        :return: True for test purpose
        """
        isOnline = self.ui.isOnline.isChecked()
        if isOnline:
            iers.conf.auto_download = True
            iers.conf.auto_max_age = 30

        else:
            iers.conf.auto_download = False
            iers.conf.auto_max_age = 99999
        return True

    def versionPackage(self, packageName):
        """
        versionPackage will look the package up in pypi.org website and parses
        the resulting versions. if you have an alpa or beta release found it
        returns the newest version for download and install.

        :param packageName:
        :return: None or the newest possible package
        """
        url = f'https://pypi.python.org/pypi/{packageName}/json'
        try:
            response = requests.get(url).json()

        except Exception as e:
            self.log.critical(f'Cannot determine package version: {e}')
            return None

        vPackage = list(response['releases'].keys())
        vPackage.sort(key=StrictVersion, reverse=True)

        verAlpha = [x for x in vPackage if 'a' in x]
        verBeta = [x for x in vPackage if 'b' in x]
        verRelease = [x for x in vPackage if 'b' not in x and 'a' not in x]

        self.log.debug(f'Package Alpha  : {verAlpha[:10]}')
        self.log.debug(f'Package Beta   : {verBeta[:10]}')
        self.log.debug(f'Package Release: {verRelease[:10]}')

        if self.ui.versionBeta.isChecked():
            vPackage = verBeta

        else:
            vPackage = verRelease

        return vPackage[0]

    def showUpdates(self):
        """
        showUpdates compares the actual installed and running package with the
        one on the server. if you have a newer version available, mw4 will inform
        the user about it.

        :return: success
        """
        packageName = 'mountwizzard4'
        actPackage = importlib_metadata.version(packageName)
        self.ui.versionActual.setText(actPackage)

        if not self.ui.isOnline.isChecked():
            self.ui.versionAvailable.setText('disabled')
            self.ui.installVersion.setEnabled(False)
            return False

        availPackage = self.versionPackage(packageName)

        if availPackage is None:
            self.app.message.emit('Failed get actual package from server', 2)
            return False

        self.ui.versionAvailable.setText(availPackage)
        self.ui.installVersion.setEnabled(True)

        if StrictVersion(availPackage) > StrictVersion(actPackage):
            self.app.message.emit('A new version of MountWizzard is available', 1)
        return True

    def isVenv(self):
        """
        detects if the actual package is running in a virtual environment. this
        should be the case in any situation as mw4 should be installed in a venv.

        :return: status
        """
        hasReal = hasattr(sys, 'real_prefix')
        hasBase = hasattr(sys, 'base_prefix')

        status = hasReal or hasBase and sys.base_prefix != sys.prefix
        self.log.debug(f'venv: [{status}], hasReal:[{hasReal}], hasBase:[{hasBase}]')
        return status

    @staticmethod
    def formatPIP(line=''):
        """
        formatPIP shortens the stdout line for presenting it to the user. as the
        lines are really long, mw4 concentrates on package names and action.

        :param line:
        :return: formatted line
        """
        if line.startswith(' '):
            return ''

        elif line.startswith('Requirement'):
            val = line.split(':')
            prefix = val[0]
            packageName = val[1].split('<')[0].split('>')[0].split('=')[0].split(' ')[1]
            line = f'{prefix} : {packageName}'

        elif line.startswith('Collecting'):
            line = line.split('<')[0].split('>')[0].split('=')[0]

        elif line.startswith('Installing') or line.startswith('Building'):
            line = line.split(':')[0]

        else:
            line = line.split('\n')[0]

        return line

    def runInstall(self, versionPackage='', timeout=60):
        """
        :param versionPackage:   package version to install
        :param timeout:
        :return: success
        """
        runnable = ['pip',
                    'install',
                    f'mountwizzard4=={versionPackage}',
                    '--disable-pip-version-check',
                    ]

        timeStart = time.time()
        try:
            self.process = subprocess.Popen(args=runnable,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.STDOUT,
                                            text=True,
                                            )
            for stdout_line in iter(self.process.stdout.readline, ""):
                line = self.formatPIP(line=stdout_line)
                if line:
                    self.app.message.emit(line, 0)

            output = self.process.communicate(timeout=timeout)[0]

        except subprocess.TimeoutExpired as e:
            self.log.critical(e)
            return False, None

        except Exception as e:
            self.log.critical(f'error: {e} happened')
            return False, None

        else:
            delta = time.time() - timeStart
            retCode = str(self.process.returncode)
            self.log.debug(f'pip install took {delta}s [{retCode}] [{output}]')

        success = (self.process.returncode == 0)
        return success, versionPackage

    def installFinished(self, result):
        """
        installFinished is called when the installation thread is finished. It
        writes the final messages to the user and resets the gui to default.

        :param result:
        :return: success
        """
        if isinstance(result, tuple):
            success, versionPackage = result
        else:
            success = False

        if success:
            self.app.message.emit(f'MountWizzard4 {versionPackage} installed', 1)
            self.app.message.emit('Please restart to enable new version', 1)
            packages = sorted(["%s==%s" % (i.key, i.version) for i in working_set])
            self.log.debug(f'After update:   {packages}')

        else:
            self.app.message.emit('Could not install update installation ', 2)

        self.mutexInstall.unlock()
        self.changeStyleDynamic(self.ui.installVersion, 'running', False)
        return success

    def installVersion(self):
        """
        installVersion updates mw4 with the standard pip package installer.
        this is actually only tested and ok for running in a virtual environment.
        updates have to run only once at a time, so a mutex ensures this. If
        everything is ok, a thread it started doing the install work and a
        callback is defined when finished.

        as observation, installation on windows side takes for some reasons
        longer than in linux or osx environment. therefore an extended timeout is
        chosen.

        :return: True for test purpose
        """
        if pConf.isWindows:
            timeout = 180

        else:
            timeout = 90

        if not self.isVenv():
            self.app.message.emit('MW4 not running in an virtual environment', 2)
            return False

        if not self.mutexInstall.tryLock():
            self.app.message.emit('Install already running', 2)
            return False

        packages = sorted(["%s==%s" % (i.key, i.version) for i in working_set])
        self.log.debug(f'Before update:  {packages}')

        versionPackage = self.ui.versionAvailable.text()
        self.changeStyleDynamic(self.ui.installVersion, 'running', True)
        self.app.message.emit(f'Installing [{versionPackage}] please wait', 1)

        worker = tpool.Worker(self.runInstall,
                              versionPackage=versionPackage,
                              timeout=timeout)

        worker.signals.result.connect(self.installFinished)
        self.threadPool.start(worker)
        return True

    def setLoggingLevel(self):
        """
        :return: nothing
        """
        if self.ui.loglevelDebugTrace.isChecked():
            setCustomLoggingLevel('TRACE')

        elif self.ui.loglevelDebug.isChecked():
            setCustomLoggingLevel('DEBUG')

        elif self.ui.loglevelStandard.isChecked():
            setCustomLoggingLevel('WARN')

    def setupAudioGui(self):
        """
        :return: True for test purpose
        """
        self.guiAudioList['MountSlew'] = self.ui.soundMountSlewFinished
        self.guiAudioList['DomeSlew'] = self.ui.soundDomeSlewFinished
        self.guiAudioList['MountAlert'] = self.ui.soundMountAlert
        self.guiAudioList['ModelingFinished'] = self.ui.soundModelingFinished
        self.guiAudioList['ImageSaved'] = self.ui.soundImageSaved
        self.guiAudioList['ImageSolved'] = self.ui.soundImageSolved

        for itemKey, itemValue in self.guiAudioList.items():
            self.guiAudioList[itemKey].addItem('None')
            self.guiAudioList[itemKey].addItem('Beep')
            self.guiAudioList[itemKey].addItem('Beep1')
            self.guiAudioList[itemKey].addItem('Beep2')
            self.guiAudioList[itemKey].addItem('Bleep')
            self.guiAudioList[itemKey].addItem('Pan1')
            self.guiAudioList[itemKey].addItem('Pan2')
            self.guiAudioList[itemKey].addItem('Horn')
            self.guiAudioList[itemKey].addItem('Alarm')
            self.guiAudioList[itemKey].addItem('Alert')
        return True

    def setupAudioSignals(self):
        """
        :return: True for test purpose
        """
        if not pConf.isAvailable:
            return False

        self.audioSignalsSet['Beep'] = ':/sound/beep.wav'
        self.audioSignalsSet['Beep1'] = ':/sound/beep1.wav'
        self.audioSignalsSet['Horn'] = ':/sound/horn.wav'
        self.audioSignalsSet['Beep2'] = ':/sound/Beep2.wav'
        self.audioSignalsSet['Bleep'] = ':/sound/Bleep.wav'
        self.audioSignalsSet['Pan1'] = ':/sound/Pan1.wav'
        self.audioSignalsSet['Pan2'] = ':/sound/Pan2.wav'
        self.audioSignalsSet['Alert'] = ':/sound/alert.wav'
        self.audioSignalsSet['Alarm'] = ':/sound/alarm.wav'
        return True

    def playSound(self, value=''):
        """
        :param value:
        :return: success
        """

        listEntry = self.guiAudioList.get(value, None)
        if listEntry is None:
            return False

        sound = listEntry.currentText()
        if sound in self.audioSignalsSet:
            PyQt5.QtMultimedia.QSound.play(self.audioSignalsSet[sound])

        else:
            return False
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import numpy as np

# external packages
from skyfield.api import Angle

# local imports

__all__ = [
    'stringToDegree',
    'stringToAngle',
    'valueToAngle',
    'valueToFloat',
    'valueToInt',
    'topoToAltAz',
    'sexagesimalizeToInt',
    'checkIsHours',
    'convertToDMS',
    'convertToHMS',
    'convertToAngle',
]

log = logging.getLogger()

# conversion from value, which is
# sDD*MM:SS.S format to decimal value
# HH:MM:SS.SS format to decimal value
# sHH:MM:SS.SS format to decimal value


def stringToDegree(value):
    """
    stringToDegree takes any form of HMS / DMS string format and tries to
    convert it to a decimal number.

    :param value:
    :return: value as decimal in degrees or None if not succeeded
    """
    if not isinstance(value, str):
        return None
    if not len(value):
        return None
    if value.count('+') > 1:
        return None
    if value.count('-') > 1:
        return None
    if value == 'E':
        return None

    value = value.replace('*', ' ')
    value = value.replace(':', ' ')
    value = value.replace('deg', ' ')
    value = value.replace('"', ' ')
    value = value.replace('\'', ' ')
    value = value.split()

    try:
        value = [float(x) for x in value]

    except Exception:
        return None

    sign = 1 if value[0] >= 0 else -1
    value[0] = abs(value[0])

    if len(value) == 3:
        value = sign * (value[0] + value[1] / 60 + value[2] / 3600)
        return value

    elif len(value) == 2:
        value = sign * (value[0] + value[1] / 60)
        return value

    else:
        return None


def stringToAngle(value, preference='degrees'):
    """
    :param value:
    :param preference:
    :return:
    """
    value = stringToDegree(value)
    if value is not None:
        if preference == 'degrees':
            value = Angle(degrees=value, preference='degrees')

        else:
            value = Angle(hours=value, preference='hours')

    return value


def valueToAngle(value, preference='degrees'):
    """

    :param value:
    :param preference:
    :return:
    """
    value = valueToFloat(value)
    if value is not None:
        if preference == 'degrees':
            value = Angle(degrees=value, preference='degrees')

        else:
            value = Angle(hours=value, preference='hours')

    return value


def valueToFloat(value):
    """
    :param value:
    :return:
    """
    if value == 'E':
        return None

    try:
        value = float(value)

    except Exception:
        value = None

    return value


def valueToInt(value):
    """
    :param value:
    :return:
    """
    try:
        value = int(value)

    except Exception:
        value = None

    return value


def topoToAltAz(ha, dec, lat):
    """
    :param ha:
    :param dec:
    :param lat:
    :return:
    """
    if lat is None:
        log.warning('lat nof defined')
        return None, None

    ha = (ha * 360 / 24 + 360.0) % 360.0
    dec = np.radians(dec)
    ha = np.radians(ha)
    lat = np.radians(lat)
    alt = np.arcsin(np.sin(dec) * np.sin(lat) + np.cos(dec) * np.cos(lat) * np.cos(ha))
    value = (np.sin(dec) - np.sin(alt) * np.sin(lat)) / (np.cos(alt) * np.cos(lat))

    value = np.clip(value, -1, 1)

    A = np.arccos(value)
    A = np.degrees(A)
    alt = np.degrees(alt)

    if np.sin(ha) >= 0.0:
        az = 360.0 - A

    else:
        az = A

    return alt, az


def sexagesimalizeToInt(value, decimals=0):
    """
    :param value:
    :param decimals:
    :return:
    """
    sign = int(np.sign(value))
    value = abs(value)
    power = 10 ** decimals
    n = int(7200 * power * value + 1) // 2
    n, fraction = divmod(n, power)
    n, seconds = divmod(n, 60)
    n, minutes = divmod(n, 60)

    return sign, n, minutes, seconds, fraction


def checkIsHours(value):
    """
    checkIsHours tries to distinguish if the value describes hours or degrees

    :param value: string
    :return:
    """
    if not isinstance(value, str):
        return False

    if '*' in value:
        return False

    if '+' in value:
        return False

    if '-' in value:
        return False

    return True


def convertToAngle(value, isHours=None):
    """
    convertToAngle ties to take any value and converts is to the right form in
    skyfield angles. if the value is a string, we convert it to float value. in
    case of isHours, we should have already a string, which represents hours
    inside. if the value is float and we expect to have hours, the float value is
    normally given in degrees, so we have to calculate the numbers in hours.

    :param value:
    :param isHours:
    :return: angle as skyfield angle
    """
    if isinstance(value, str):
        value = stringToDegree(value)
        if value is None:
            return None

    else:
        if isHours:
            value *= 24 / 360

    if isHours:
        angle = Angle(hours=value)

    else:
        angle = Angle(degrees=value)

    return angle


def convertToDMS(dec):
    """
    takes the given DEC value, which should be in DMS format (but different
    types) and convert it to solve-field readable string in sDD:MM:SS

    :param dec: declination as Angle
    :return: converted value as string
    """
    if isinstance(dec, (float, int)):
        dec = Angle(degrees=dec)

    if isinstance(dec, str):
        return ''

    t = Angle.signed_dms(dec)
    sign = '+' if dec.degrees > 0 else '-'
    value = f'{sign}{t[1]:02.0f}:{t[2]:02.0f}:{t[3]:02.0f}'
    return value


def convertToHMS(ra):
    """
    takes the given RA value, which should be in HMS format (but different
    types) and convert it to solve-field readable string in HH:MM:SS

    :param ra: right ascension as Angle
    :return: converted value as string
    """
    if isinstance(ra, (float, int)):
        ra = Angle(hours=ra)

    if isinstance(ra, str):
        return ''

    t = Angle.signed_hms(ra)
    value = f'{t[1]:02.0f}:{t[2]:02.0f}:{t[3]:02.0f}'

    return value



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import logging

# external packages
from skyfield.api import Star, Angle
import numpy

# local imports
from mountcontrol.convert import stringToDegree
from mountcontrol.convert import valueToAngle
from mountcontrol.convert import valueToFloat
from mountcontrol.convert import valueToInt
from mountcontrol.convert import topoToAltAz


class ModelStar(object):
    """
    The class ModelStar inherits all information and handling of one star in
    the alignment model used by the mount and the data in the mount and provides
    the abstracted interface to a 10 micron mount.
    The coordinates are in JNow topocentric

        >>> settings = ModelStar(
        >>>                     coord=None,
        >>>                     errorRMS=0,
        >>>                     errorAngle=0,
        >>>                     number=0,
        >>>                     obsSite=None,
        >>>                     )

    point could be from type skyfield.api.Star or just a tuple of (ha, dec) where
    the format should be float or the 10micron string format.

    Command protocol (from2.8.15 onwards):
    "HH:MM:SS.SS,+dd*mm:ss.s,eeee.e,ppp#" where HH:MM:SS.SS is the hour angle
    of the alignment star in hours, minutes, seconds and hundredths of second
    (from 0h to 23h59m59.99s), +dd*mm:ss.s is the declination of the alignment
    star in degrees, arcminutes, arcseconds and tenths of arcsecond, eeee.e is
    the error between the star and the alignment model in arcseconds, ppp is the
    polar angle of the measured star with respect to the modeled star in the
    equatorial system in degrees from 0 to 359 (0 towards the north pole,
    90 towards east).
    """

    __all__ = ['ModelStar',
               ]

    log = logging.getLogger(__name__)

    def __init__(self,
                 coord=None,
                 errorRMS=None,
                 errorAngle=None,
                 number=None,
                 obsSite=None,
                 ):

        self.obsSite = obsSite
        self.coord = coord
        self.errorRMS = errorRMS
        self.errorAngle = errorAngle
        self.number = number

    @property
    def coord(self):
        return self._coord

    @coord.setter
    def coord(self, value):
        if not isinstance(value, (tuple, list, Star)):
            self._coord = None
            return

        if isinstance(value, Star):
            self._coord = value
            ha = self._coord.ra.hours
            dec = self._coord.dec.degrees

        else:
            if len(value) != 2:
                self._coord = None
                return

            ha, dec = value
            ha = stringToDegree(ha)
            dec = stringToDegree(dec)

            if not ha or not dec:
                self._coord = None
                self.log.warning('Malformed value: {0}'.format(value))
                return
            self._coord = Star(ra_hours=ha, dec_degrees=dec)

        loc = self.obsSite.location
        if loc is None:
            return

        lat = loc.latitude.degrees
        alt, az = topoToAltAz(ha, dec, lat)
        self._alt = Angle(degrees=alt)
        self._az = Angle(degrees=az)

    @property
    def number(self):
        return self._number

    @number.setter
    def number(self, value):
        self._number = valueToInt(value)

    @property
    def errorRMS(self):
        return self._errorRMS

    @errorRMS.setter
    def errorRMS(self, value):
        self._errorRMS = valueToFloat(value)

    @property
    def errorAngle(self):
        return self._errorAngle

    @errorAngle.setter
    def errorAngle(self, value):
        if isinstance(value, Angle):
            self._errorAngle = value

        else:
            self._errorAngle = valueToAngle(value)

    @property
    def alt(self):
        return self._alt

    @alt.setter
    def alt(self, value):
        self._alt = valueToFloat(value)

    @property
    def az(self):
        return self._az

    @az.setter
    def az(self, value):
        self._az = valueToFloat(value)

    def errorRA(self):
        if self._errorRMS is not None and self._errorAngle is not None:
            return self._errorRMS * numpy.sin(self._errorAngle.radians)
        else:
            return None

    def errorDEC(self):
        if self._errorRMS is not None and self._errorAngle is not None:
            return self._errorRMS * numpy.cos(self._errorAngle.radians)
        else:
            return None

    def __gt__(self, other):
        if other > self._errorRMS:
            return True
        else:
            return False

    def __ge__(self, other):
        if other >= self._errorRMS:
            return True
        else:
            return False

    def __lt__(self, other):
        if other < self._errorRMS:
            return True
        else:
            return False

    def __le__(self, other):
        if other <= self._errorRMS:
            return True
        else:
            return False

    def __eq__(self, other):
        if other == self._errorRMS:
            return True
        else:
            return False



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import sys

# external packages
from PyQt5.QtCore import QObject, pyqtSignal, QRunnable, pyqtSlot

# local imports


__all__ = ['Worker',
           ]


class WorkerSignals(QObject):
    """
    The WorkerSignals class offers a list of signals to be used and instantiated by the Worker
    class to get signals for error, finished and result to be transferred to the caller back
    """

    __all__ = ['WorkerSignals']

    finished = pyqtSignal()
    error = pyqtSignal(object)
    result = pyqtSignal(object)


class Worker(QRunnable):
    """
    The Worker class offers a generic interface to allow any function to be executed as
    a thread in an threadpool
    """

    __all__ = ['Worker']

    log = logging.getLogger(__name__)

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()

        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

        # the worker signal must not be a class variable, but instance otherwise
        # we get trouble when having multiple threads running
        self.signals = WorkerSignals()

    @pyqtSlot()
    def run(self):
        """
        runs an arbitrary methods with it's parameters and catches the result

        :return: nothing, but sends results and status as signals
        """

        try:
            result = self.fn(*self.args, **self.kwargs)

        except Exception:
            # as we want to send a clear message to the log file
            exc_type, exc_value, exc_traceback = sys.exc_info()
            tb = exc_traceback

            # moving toward the end of the trace
            while tb.tb_next is not None:
                tb = tb.tb_next

            # getting data out for processing
            file = tb.tb_frame.f_code.co_filename
            line = tb.tb_frame.f_lineno

            errorString = f'{file}, line {line} {exc_value}'
            self.log.critical(errorString)
            self.signals.error.emit(errorString)

        else:
            self.signals.result.emit(result)

        finally:
            self.signals.finished.emit()



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import platform

# external packages
import PyQt5

# local imports
from logic.dome.domeIndi import DomeIndi
from logic.dome.domeAlpaca import DomeAlpaca
if platform.system() == 'Windows':
    from logic.dome.domeAscom import DomeAscom


class DomeSignals(PyQt5.QtCore.QObject):
    """
    The DomeSignals class offers a list of signals to be used and instantiated by
    the Mount class to get signals for triggers for finished tasks to
    enable a gui to update their values transferred to the caller back.

    This has to be done in a separate class as the signals have to be subclassed from
    QObject and the Mount class itself is subclassed from object
    """

    __all__ = ['DomeSignals']

    azimuth = PyQt5.QtCore.pyqtSignal(object)
    slewFinished = PyQt5.QtCore.pyqtSignal()
    message = PyQt5.QtCore.pyqtSignal(object)

    serverConnected = PyQt5.QtCore.pyqtSignal()
    serverDisconnected = PyQt5.QtCore.pyqtSignal(object)
    deviceConnected = PyQt5.QtCore.pyqtSignal(str)
    deviceDisconnected = PyQt5.QtCore.pyqtSignal(str)


class Dome:

    __all__ = ['Dome',
               ]

    log = logging.getLogger(__name__)

    def __init__(self, app):
        self.app = app
        self.threadPool = app.threadPool
        self.signals = DomeSignals()

        self.data = {
            'Slewing': False,
        }
        self.defaultConfig = {'framework': '',
                              'frameworks': {}}
        self.framework = ''
        self.run = {
            'indi': DomeIndi(self.app, self.signals, self.data),
            'alpaca': DomeAlpaca(self.app, self.signals, self.data),
        }

        if platform.system() == 'Windows':
            self.run['ascom'] = DomeAscom(self.app, self.signals, self.data)
            ascomSignals = self.run['ascom'].ascomSignals
            ascomSignals.serverConnected.connect(self.signals.serverConnected)
            ascomSignals.serverDisconnected.connect(self.signals.serverDisconnected)
            ascomSignals.deviceConnected.connect(self.signals.deviceConnected)
            ascomSignals.deviceDisconnected.connect(self.signals.deviceDisconnected)

        for fw in self.run:
            self.defaultConfig['frameworks'].update(self.run[fw].defaultConfig)

        # signalling from subclasses to main
        alpacaSignals = self.run['alpaca'].client.signals
        alpacaSignals.serverConnected.connect(self.signals.serverConnected)
        alpacaSignals.serverDisconnected.connect(self.signals.serverDisconnected)
        alpacaSignals.deviceConnected.connect(self.signals.deviceConnected)
        alpacaSignals.deviceDisconnected.connect(self.signals.deviceDisconnected)

        indiSignals = self.run['indi'].client.signals
        indiSignals.serverConnected.connect(self.signals.serverConnected)
        indiSignals.serverDisconnected.connect(self.signals.serverDisconnected)
        indiSignals.deviceConnected.connect(self.signals.deviceConnected)
        indiSignals.deviceDisconnected.connect(self.signals.deviceDisconnected)

        self.app.update1s.connect(self.checkSlewingDome)
        self.useGeometry = False
        self.isSlewing = False
        self.counterStartSlewing = -1
        self.settlingTime = 0
        self.settlingWait = PyQt5.QtCore.QTimer()
        self.settlingWait.setSingleShot(True)
        self.settlingWait.timeout.connect(self.waitSettlingAndEmit)

    def startCommunication(self, loadConfig=False):
        """
        :param loadConfig:
        :return:
        """
        if self.framework not in self.run.keys():
            return False

        suc = self.run[self.framework].startCommunication(loadConfig=loadConfig)
        return suc

    def stopCommunication(self):
        """
        :return:
        """
        if self.framework not in self.run.keys():
            return False

        self.signals.message.emit('')
        suc = self.run[self.framework].stopCommunication()
        return suc

    def waitSettlingAndEmit(self):
        """
        :return: true for test purpose
        """

        self.signals.slewFinished.emit()
        self.signals.message.emit('')

        return True

    def checkSlewingDome(self):
        """
        :return:
        """
        if self.isSlewing:
            self.signals.message.emit('slewing')
            self.counterStartSlewing = -1
            if not self.data.get('Slewing', True):
                self.isSlewing = False
                self.signals.message.emit('wait settle')
                self.settlingWait.start(self.settlingTime * 1000)

        else:
            if self.data.get('Slewing', True):
                self.log.debug('Slewing start by signal')
                self.isSlewing = True

            else:
                if self.counterStartSlewing == 0:
                    self.log.debug('Slewing start by counter')
                    self.isSlewing = True
                self.counterStartSlewing -= 1

        return True

    def slewDome(self, altitude=0, azimuth=0):
        """
        :param altitude:
        :param azimuth:
        :return: success
        """
        if not self.data:
            self.log.error('No data dict available')
            return False

        mount = self.app.mount
        if self.useGeometry:
            alt, az, _, _, _ = mount.calcTransformationMatrices()

            if alt is None or az is None:
                self.log.info(f'Geometry error, alt:{altitude}, az:{azimuth}')
                alt = altitude
                az = azimuth

            else:
                alt = alt.degrees
                az = az.degrees

        else:
            alt = altitude
            az = azimuth

        self.signals.message.emit('slewing')
        self.counterStartSlewing = 3
        self.run[self.framework].slewToAltAz(azimuth=az, altitude=alt)
        delta = azimuth - az
        return delta

    def openShutter(self):
        """
        :return: success
        """
        if not self.data:
            self.log.error('No data dict available')
            return False

        suc = self.run[self.framework].openShutter()
        return suc

    def closeShutter(self):
        """
        :return: success
        """
        if not self.data:
            self.log.error('No data dict available')
            return False

        suc = self.run[self.framework].closeShutter()
        return suc

    def abortSlew(self):
        """
        :return: success
        """
        if not self.data:
            self.log.error('No data dict available')
            return False

        suc = self.run[self.framework].abortSlew()
        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging
import os
import json
from queue import Queue
import platform

# external packages
from PyQt5.QtCore import QObject, pyqtSignal, QThreadPool, QTimer
from skyfield.api import Topos
from importlib_metadata import version

# local import
from base.loggerMW import setCustomLoggingLevel
from mountcontrol import qtmount
from gui.mainWindow.mainW import MainWindow
from logic.powerswitch.kmRelay import KMRelay
from logic.modeldata.buildpoints import DataPoint
from logic.modeldata.hipparcos import Hipparcos
from logic.dome.dome import Dome
from logic.imaging.camera import Camera
from logic.imaging.filter import Filter
from logic.imaging.focuser import Focuser
from logic.environment.sensorWeather import SensorWeather
from logic.environment.skymeter import Skymeter
from logic.environment.onlineWeather import OnlineWeather
from logic.environment.directWeather import DirectWeather
from logic.environment.weatherUPB import WeatherUPB
from logic.cover.cover import Cover
from logic.telescope.telescope import Telescope
from logic.powerswitch.pegasusUPB import PegasusUPB
from logic.measure.measure import MeasureData
from logic.remote.remote import Remote
from logic.astrometry.astrometry import Astrometry


class MountWizzard4(QObject):
    """
    MountWizzard4 class is the main class for the application. it loads all
    windows and classes needed to fulfil the work of mountwizzard. any gui work
    should be handled through the window classes. main class is for setup,
    config, start, persist and shutdown the application.
    """

    __all__ = ['MountWizzard4']
    __version__ = version('mountwizzard4')

    log = logging.getLogger(__name__)

    message = pyqtSignal(str, int)
    messageQueue = Queue()
    redrawHemisphere = pyqtSignal()
    updatePointMarker = pyqtSignal()
    drawBuildPoints = pyqtSignal()
    drawHorizonPoints = pyqtSignal()
    updateDomeSettings = pyqtSignal()
    sendSatelliteData = pyqtSignal()
    showImage = pyqtSignal(str)
    showAnalyse = pyqtSignal(str)
    remoteCommand = pyqtSignal(str)

    update0_1s = pyqtSignal()
    update1s = pyqtSignal()
    update3s = pyqtSignal()
    update10s = pyqtSignal()
    update60s = pyqtSignal()
    update3m = pyqtSignal()
    update10m = pyqtSignal()
    update30m = pyqtSignal()
    update1h = pyqtSignal()

    def __init__(self,
                 mwGlob=None,
                 application=None
                 ):

        super().__init__()

        self.application = application
        self.expireData = False
        self.mountUp = False
        self.mwGlob = mwGlob
        self.timerCounter = 0
        self.mainW = None
        self.threadPool = QThreadPool()
        self.threadPool.setMaxThreadCount(20)
        self.message.connect(self.writeMessageQueue)
        self.config = {}
        self.loadConfig()
        self.deviceStat = {
            'dome': None,
            'mount': None,
            'camera': None,
            'astrometry': None,
            'environOverall': None,
            'sensorWeather': None,
            'directWeather': None,
            'onlineWeather': None,
            'powerWeather': None,
            'skymeter': None,
            'cover': None,
            'telescope': None,
            'power': None,
            'remote': None,
            'relay': None,
            'measure': None,
        }
        profile = self.config.get('profileName', '-')
        self.messageQueue.put(('MountWizzard4 started', 1))
        self.messageQueue.put((f'Workdir is: [{self.mwGlob["workDir"]}]', 1))
        self.messageQueue.put((f'Profile              [{profile}] loaded', 1))

        # initialize commands to mount
        pathToData = self.mwGlob['dataDir']
        self.mount = qtmount.Mount(host='localhost',
                                   MAC='00.c0.08.87.35.db',
                                   threadPool=self.threadPool,
                                   pathToData=pathToData,
                                   verbose=False,
                                   )

        # setting location to last know config
        topo = self.initConfig()
        self.mount.obsSite.location = topo
        self.mount.signals.mountUp.connect(self.loadMountData)
        self.ephemeris = self.mount.obsSite.loader('de421_23.bsp')
        self.relay = KMRelay()
        self.sensorWeather = SensorWeather(self)
        self.onlineWeather = OnlineWeather(self)
        self.directWeather = DirectWeather(self)
        self.powerWeather = WeatherUPB(self)
        self.cover = Cover(self)
        self.dome = Dome(self)
        self.camera = Camera(self)
        self.filter = Filter(self)
        self.focuser = Focuser(self)
        self.telescope = Telescope(self)
        self.skymeter = Skymeter(self)
        self.power = PegasusUPB(self)
        self.data = DataPoint(self)
        self.hipparcos = Hipparcos(self)
        self.measure = MeasureData(self)
        self.remote = Remote(self)
        self.astrometry = Astrometry(self)
        self.automation = self.checkAndSetAutomation()

        self.uiWindows = {}
        self.mainW = MainWindow(self)

        self.mount.startTimers()
        self.timer0_1s = QTimer()
        self.timer0_1s.setSingleShot(False)
        self.timer0_1s.timeout.connect(self.sendUpdate)
        self.timer0_1s.start(100)
        self.application.aboutToQuit.connect(self.aboutToQuit)

        if os.path.isfile('test.txt'):
            self.update3s.connect(self.quit)

    def checkAndSetAutomation(self):
        """
        the windows automation with pywinauto has a serious bug in python lib.
        the bugfix is done from python 3.8.2 onwards. so to enable this work,
        we have to check the python version used and set the topic adequately.

        :return:
        """
        if platform.system() != 'Windows':
            return None
        if platform.python_version() < '3.8.2':
            return None

        from logic.automation.automateWindows import AutomateWindows
        automation = AutomateWindows(self)

        return automation

    def initConfig(self):
        """
        :return: topo object with location
        """
        lat = self.config.get('topoLat', 51.47)
        lon = self.config.get('topoLon', 0)
        elev = self.config.get('topoElev', 46)
        topo = Topos(longitude_degrees=lon,
                     latitude_degrees=lat,
                     elevation_m=elev)

        config = self.config.get('mainW', {})
        if config.get('loglevelDeepDebug', False):
            level = 'DEBUG'
        elif config.get('loglevelDebug', False):
            level = 'INFO'
        else:
            level = 'WARN'
        setCustomLoggingLevel(level)
        return topo

    def storeConfig(self):
        """
        :return: success for test purpose
        """
        location = self.mount.obsSite.location
        if location is not None:
            self.config['topoLat'] = location.latitude.degrees
            self.config['topoLon'] = location.longitude.degrees
            self.config['topoElev'] = location.elevation.m
        return True

    def sendUpdate(self):
        """
        sendUpdate send regular signals in 1 and 10 seconds to enable regular
        tasks. it tries to avoid sending the signals at the same time.

        :return: true for test purpose
        """
        self.timerCounter += 1
        if self.timerCounter % 1 == 0:
            self.update0_1s.emit()
        if (self.timerCounter + 5) % 10 == 0:
            self.update1s.emit()
        if (self.timerCounter + 10) % 30 == 0:
            self.update3s.emit()
        if (self.timerCounter + 20) % 100 == 0:
            self.update10s.emit()
        if (self.timerCounter + 25) % 600 == 0:
            self.update60s.emit()
        if (self.timerCounter + 12) % 1800 == 0:
            self.update3m.emit()
        if (self.timerCounter + 13) % 6000 == 0:
            self.update10m.emit()
        if (self.timerCounter + 14) % 18000 == 0:
            self.update30m.emit()
        if (self.timerCounter + 15) % 36000 == 0:
            self.update1h.emit()
        return True

    def aboutToQuit(self):
        """
        :return:    True for test purpose
        """
        self.timer0_1s.stop()
        self.mount.stopTimers()
        return True

    def quit(self):
        """
        :return:    True for test purpose
        """
        self.aboutToQuit()
        self.mount.mountUp = False
        self.threadPool.waitForDone(5000)
        self.message.emit('MountWizzard4 manual stopped', 1)
        self.application.quit()
        return True

    @staticmethod
    def defaultConfig(config=None):
        """
        :param config:
        :return:
        """
        if config is None:
            config = dict()

        config['profileName'] = 'config'
        config['version'] = '4.0'
        return config

    def loadConfig(self, name=None):
        """
        loadConfig loads a json file from disk and stores it to the config
        dicts for persistent data. if a file path is given, that's the relevant
        file, otherwise loadConfig loads from th default file, which is
        config.cfg

        :param      name:   name of the config file
        :return:    success if file could be loaded
        """
        configDir = self.mwGlob['configDir']
        if name is None:
            profileFile = f'{configDir}/profile'
            if os.path.isfile(profileFile):
                with open(profileFile, 'r') as profile:
                    name = profile.readline().strip()
            else:
                name = 'config'

        self.config['profileName'] = name

        fileName = f'{configDir}/{name}.cfg'

        if not os.path.isfile(fileName):
            self.config = self.defaultConfig()
            self.log.info(f'Config file {fileName} not existing')
            return False

        try:
            with open(fileName, 'r') as configFile:
                configData = json.load(configFile)
        except Exception as e:
            self.log.critical(f'Cannot parse: {fileName}, error: {e}')
            self.config = self.defaultConfig()
            return False
        else:
            self.config = configData

        return True

    @staticmethod
    def convertData(data):
        """
        convertDate tries to convert data from an older or newer version of the
        config file to the actual needed one.

        :param      data: config data as dict
        :return:    data: config data as dict
        """
        return data

    def saveConfig(self, name=None):
        """
        :param      name:   name of the config file
        :return:    success
        """
        configDir = self.mwGlob['configDir']
        if name is None:
            name = self.config['profileName']

        profileFile = f'{configDir}/profile'
        with open(profileFile, 'w') as profile:
            profile.writelines(f'{name}')

        fileName = configDir + '/' + name + '.cfg'
        with open(fileName, 'w') as outfile:
            json.dump(self.config,
                      outfile,
                      sort_keys=True,
                      indent=4)
        return True

    def loadMountData(self, status):
        """
        :param      status: connection status to mount computer
        :return:    status how it was called
        """
        if status and not self.mountUp:
            self.mount.getFW()
            self.mount.getLocation()
            self.mount.cycleSetting()
            self.mainW.refreshName()
            self.mainW.refreshModel()
            self.mount.getTLE()
            self.mountUp = True
            return True

        elif not status and self.mountUp:
            location = self.mount.obsSite.location
            self.mount.resetData()
            self.mount.obsSite.location = location
            self.mountUp = False
            return False
        return status

    def writeMessageQueue(self, message, mType):
        """
        :param message:
        :param mType:
        :return: True for test purpose
        """
        self.log.ui(f'Message window: [{message}]')
        self.messageQueue.put((message, mType))
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages
import numpy as np
import matplotlib.path as mpath

# local import


class HemisphereWindowExt:
    """
    the HemisphereWindowExt window class handles

    """

    __all__ = ['HemisphereWindowExt',
               ]

    @staticmethod
    def markerPoint():
        """
        markerPoint constructs a custom marker for presentation of modeldata points by
        concatenating the circle with an internal cutout of the star

        :return: marker
        """

        circleB = mpath.Path.unit_circle()
        circleS = mpath.Path.unit_circle()
        verts = np.concatenate([circleB.vertices, 0.5 * circleS.vertices])
        codes = np.concatenate([circleB.codes, circleS.codes])
        marker = mpath.Path(verts, codes)

        return marker

    @staticmethod
    def markerAltAz():
        """
        markerAltAz constructs a custom marker for AltAz pointer by
        concatenating the circle with an internal cutout of the star

        :return: marker
        """

        circleB = mpath.Path.unit_circle()
        circleM = mpath.Path.unit_circle()
        circleS = mpath.Path.unit_circle()
        circleC = mpath.Path.unit_circle()
        verts = np.concatenate([circleB.vertices,
                                0.8 * circleM.vertices,
                                0.15 * circleS.vertices,
                                0.1 * circleC.vertices])
        codes = np.concatenate([circleB.codes,
                                circleM.codes,
                                circleS.codes,
                                circleC.codes])
        marker = mpath.Path(verts, codes)

        return marker

    @staticmethod
    def markerStar():
        """
        markerStar constructs a custom marker for presentation of modeldata points by
        concatenating the circle with an internal cutout of the star

        :return: marker
        """

        star = mpath.Path.unit_regular_star(8)
        verts = np.concatenate([star.vertices])
        codes = np.concatenate([star.codes])
        marker = mpath.Path(verts, codes)

        return marker

    def setOperationMode(self):
        """
        setOperationMode changes the operation mode of the hemisphere window(s) depending
        on the choice, colors and styles will be changed.

        :return: success
        """

        if self.ui.checkEditNone.isChecked():
            self.operationMode = 'normal'
            self.ui.addPositionToHorizon.setEnabled(False)

        elif self.ui.checkEditBuildPoints.isChecked():
            self.operationMode = 'build'
            self.ui.addPositionToHorizon.setEnabled(False)

        elif self.ui.checkEditHorizonMask.isChecked():
            self.operationMode = 'horizon'
            self.ui.addPositionToHorizon.setEnabled(True)

        elif self.ui.checkPolarAlignment.isChecked():
            self.ui.checkShowAlignStar.setChecked(True)
            self.operationMode = 'star'
            self.ui.addPositionToHorizon.setEnabled(False)

        self.drawHemisphere()

        return True

    def showMouseCoordinates(self, event):
        """

        :param event:
        :return: success
        """

        if event.xdata is None:
            return False

        if event.ydata is None:
            return False

        self.ui.altitude.setText(f'{event.ydata:3.1f}')
        self.ui.azimuth.setText(f'{event.xdata:3.1f}')

        return True

    def slewSelectedTarget(self, alignType='normal'):
        """

        :param alignType:
        :return: success
        """

        azimuthT = self.app.mount.obsSite.AzTarget.degrees
        altitudeT = self.app.mount.obsSite.AltTarget.degrees

        if self.app.deviceStat['dome']:
            delta = self.app.dome.slewDome(altitude=altitudeT,
                                           azimuth=azimuthT)

            geoStat = 'Geometry corrected' if delta else 'Equal mount'
            text = f'Slewing dome:        {geoStat}, az: {azimuthT:3.1f} delta: {delta:3.1f}'
            self.app.message.emit(text, 0)

        suc = self.app.mount.obsSite.startSlewing(slewType='keep')

        if suc:
            self.app.message.emit('Slewing mount', 0)

        else:
            self.app.message.emit('Cannot slew to: {0}, {1}'.format(azimuthT, altitudeT), 2)

        return suc

    def onMouseNormal(self, event):
        """
        onMouseNormal handles the mouse event in normal mode. this means only a double
        click is possible and offers the opportunity to slew the telescope to a certain
        position in sky selected by the mouse.

        :param event: mouse events
        :return: success
        """

        if not event.inaxes:
            return False

        if event.button != 1 or not event.dblclick:
            return False

        azimuth = int(event.xdata + 0.5)
        altitude = int(event.ydata + 0.5)

        textFormat = 'Do you want to slew the mount to:\n\nAzimuth:\t{0}°\nAltitude:\t{1}°'
        question = textFormat.format(azimuth, altitude)

        suc = self.messageDialog(self, 'Slewing mount', question)

        if not suc:
            return False

        suc = self.app.mount.obsSite.setTargetAltAz(alt_degrees=altitude,
                                                    az_degrees=azimuth)
        if not suc:
            self.app.message.emit('Cannot slew to: {0}, {1}'.format(azimuth, altitude), 2)
            return False

        suc = self.slewSelectedTarget(alignType='keep')

        return suc

    def addHorizonPointManual(self):
        """
        :return:
        """
        data = self.app.data
        alt = self.app.mount.obsSite.Alt
        az = self.app.mount.obsSite.Az

        if alt is None or az is None:
            return False

        index = self.getIndexPointX(x=az.degrees, plane=data.horizonP)
        if index is None and data.horizonP:
            return False

        suc = data.addHorizonP(value=(alt.degrees, az.degrees), position=index)
        if suc:
            self.drawHemisphere()
        return suc

    def addHorizonPoint(self, data=None, event=None):
        """
        addHorizonPoint calculates from the position of the left mouse click the position
        where the next horizon point should be added. the coordinates are given from mouse
        click itself.

        :param data: point in tuples (alt, az)
        :param event: mouse event
        :return:
        """
        index = self.getIndexPointX(x=event.xdata, plane=data.horizonP)
        if index is None and data.horizonP:
            return False

        suc = data.addHorizonP(value=(event.ydata, event.xdata),
                               position=index)
        return suc

    def deleteHorizonPoint(self, data=None, event=None):
        """
        deleteHorizonPoint selects the next horizon point in distance max and tries to
        delete it. there have to be at least 2 horizon point left.

        :param data: point in tuples (alt, az)
        :param event: mouse event
        :return: success
        """
        index = self.getIndexPoint(event=event, plane=data.horizonP)

        if index is None:
            return False

        suc = False
        if len(data.horizonP) > 0:
            suc = data.delHorizonP(position=index)

        return suc

    def editHorizonMask(self, data=None, event=None):
        """
        editHorizonMask does dispatching the different mouse clicks for adding or deleting
        horizon mask points and call the function accordingly.

        :param data: point in tuples (alt, az)
        :param event: mouse event
        :return: success
        """
        if event.button == 1:
            suc = self.addHorizonPoint(data=data, event=event)

        elif event.button == 3:
            suc = self.deleteHorizonPoint(data=data, event=event)

        else:
            return False

        if data.horizonP is None:
            return False

        self.drawHemisphere()
        return suc

    def addBuildPoint(self, data=None, event=None):
        """
        addBuildPoint calculates from the position of the left mouse click the position
        where the next modeldata point should be added. the coordinates are given from mouse
        click itself.

        :param data: point in tuples (alt, az)
        :param event: mouse event
        :return:
        """
        index = self.getIndexPoint(event=event, plane=data.buildP, epsilon=360)
        if index is None:
            return False

        index += 1
        suc = data.addBuildP(value=(event.ydata, event.xdata, True),
                             position=index)
        if not suc:
            return False

        return True

    def deleteBuildPoint(self, data=None, event=None):
        """
        deleteBuildPoint selects the next modeldata point in distance max and tries to
        delete it. there have to be at least 2 horizon point left.

        :param data: point in tuples (alt, az)
        :param event: mouse event
        :return: success
        """
        index = self.getIndexPoint(event=event, plane=data.buildP)
        if index is None:
            return False

        suc = data.delBuildP(position=index)

        return suc

    def editBuildPoints(self, data=None, event=None):
        """
        editBuildPoints does dispatching the different mouse clicks for adding or deleting
        model data points and call the function accordingly.

        :param data: points in tuples (alt, az)
        :param event: mouse event
        :return: success
        """

        if event.button == 1:
            suc = self.addBuildPoint(data=data, event=event)

        elif event.button == 3:
            suc = self.deleteBuildPoint(data=data, event=event)

        else:
            return False

        self.drawHemisphere()

        return suc

    def onMouseEdit(self, event):
        """
        onMouseEdit handles the mouse event in normal mode. this means depending on the
        edit mode (horizon or model points) a left click adds a new point and right click
        deletes the selected point.

        :param event: mouse events
        :return: success
        """
        data = self.app.data

        if not event.inaxes:
            return False

        if event.dblclick:
            return False

        if self.ui.checkEditHorizonMask.isChecked():
            suc = self.editHorizonMask(event=event, data=data)

        elif self.ui.checkEditBuildPoints.isChecked():
            suc = self.editBuildPoints(event=event, data=data)

        else:
            return False

        return suc

    def onMouseStar(self, event):
        """
        onMouseStar handles the mouse event in polar align mode. this means only a right
        click is possible and offers the opportunity to slew the telescope to the selected
        star and start manual polar alignment.

        :param event: mouse events
        :return: success
        """
        if not event.inaxes:
            return False

        if event.button == 1 and not event.dblclick:
            alignType = 'polar'

        elif event.button == 3 and not event.dblclick:
            alignType = 'ortho'

        else:
            return False

        hip = self.app.hipparcos
        plane = list(zip(hip.alt, hip.az))
        index = self.getIndexPoint(event=event, plane=plane, epsilon=2)
        if index is None:
            return False

        name = hip.name[index]
        ra, dec = hip.getAlignStarRaDecFromName(hip.name[index])

        textFormat = 'Align: {0}\nDo you want to slew the mount to:\n\n{1}'
        question = textFormat.format(alignType, name)

        suc = self.messageDialog(self, 'Slewing mount', question)

        if not suc:
            return False

        suc = self.app.mount.obsSite.setTargetRaDec(ra_hours=ra,
                                                    dec_degrees=dec)

        if not suc:
            self.app.message.emit(f'Cannot slew to: {name}', 2)
            return False

        suc = self.slewSelectedTarget(alignType=alignType)

        return suc

    def onMouseDispatcher(self, event):
        """
        onMouseDispatcher dispatches the button events depending on the actual operation
        mode.

        :param event: button event for parsing
        :return: True for test purpose
        """
        if self.ui.checkEditNone.isChecked():
            self.onMouseNormal(event)

        elif self.ui.checkEditBuildPoints.isChecked():
            self.onMouseEdit(event)

        elif self.ui.checkEditHorizonMask.isChecked():
            self.onMouseEdit(event)

        elif self.ui.checkPolarAlignment.isChecked():
            self.onMouseStar(event)

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages
import PyQt5
from mountcontrol.convert import valueToInt

# local import


class Power(object):
    """
    """

    def __init__(self):
        self.powerOnOFF = {'1': self.ui.powerPort1,
                           '2': self.ui.powerPort2,
                           '3': self.ui.powerPort3,
                           '4': self.ui.powerPort4,
                           }
        self.powerBoot = {'1': self.ui.powerBootPort1,
                          '2': self.ui.powerBootPort2,
                          '3': self.ui.powerBootPort3,
                          '4': self.ui.powerBootPort4,
                          }
        self.dew = {'A': self.ui.dewA,
                    'B': self.ui.dewB,
                    'C': self.ui.dewC,
                    }
        self.current = {'1': self.ui.powerCurrent1,
                        '2': self.ui.powerCurrent2,
                        '3': self.ui.powerCurrent3,
                        '4': self.ui.powerCurrent4,
                        }
        self.label = {'1': self.ui.powerLabel1,
                      '2': self.ui.powerLabel2,
                      '3': self.ui.powerLabel3,
                      '4': self.ui.powerLabel4,
                      }
        self.portUSB = {'1': self.ui.portUSB1,
                        '2': self.ui.portUSB2,
                        '3': self.ui.portUSB3,
                        '4': self.ui.portUSB4,
                        '5': self.ui.portUSB5,
                        '6': self.ui.portUSB6,
                        }

        # gui tasks
        self.ui.hubUSB.clicked.connect(self.toggleHubUSB)
        self.ui.autoDew.clicked.connect(self.toggleAutoDew)
        self.ui.rebootUPB.clicked.connect(self.rebootUPB)
        self.clickable(self.ui.adjustableOutput).connect(self.setAdjustableOutput)

        # setting gui elements
        for name, button in self.dew.items():
            self.clickable(button).connect(self.setDew)
        for name, button in self.powerOnOFF.items():
            button.clicked.connect(self.togglePowerPort)
        for name, button in self.powerBoot.items():
            button.clicked.connect(self.togglePowerBootPort)
        for name, button in self.portUSB.items():
            button.clicked.connect(self.togglePortUSB)

        # functional signals
        self.app.power.signals.version.connect(self.setGuiVersion)

        # cyclic tasks
        self.app.update1s.connect(self.updatePowerGui)

    def clearPowerGui(self):
        """
        clearPowerGui changes the state of the Pegasus values to '-'

        :return: success for test
        """

        self.ui.powerCurrent1.setText('-')
        self.ui.powerCurrent2.setText('-')
        self.ui.powerCurrent3.setText('-')
        self.ui.powerCurrent4.setText('-')
        self.ui.consumptionAvgAmps.setText('-')
        self.ui.consumptionAmpHours.setText('-')
        self.ui.consumptionWattHours.setText('-')
        self.ui.sensorVoltage.setText('-')
        self.ui.sensorCurrent.setText('-')
        self.ui.sensorPower.setText('-')
        self.ui.dewCurrentA.setText('-')
        self.ui.dewCurrentB.setText('-')

        return True

    def setGuiVersion(self, version=1):
        """
        setGuiVersion enables and disables the gui elements according to the recognized
        version the UPB.

        :param version:
        :return:
        """

        if version == 1:
            self.ui.groupDewC.setVisible(False)
            self.ui.groupPortUSB.setVisible(False)
            self.ui.groupHubUSB.setVisible(True)
            self.ui.groupAdjustableOutput.setVisible(False)

        elif version == 2:
            self.ui.groupDewC.setVisible(True)
            self.ui.groupPortUSB.setVisible(True)
            self.ui.groupHubUSB.setVisible(False)
            self.ui.groupAdjustableOutput.setVisible(True)

    def updatePowerGui(self):
        """
        updatePowerGui changes the style of the button related to the state of the Pegasus

        :return: success for test
        """

        for name, button in self.powerOnOFF.items():
            value = self.app.power.data.get(f'POWER_CONTROL.POWER_CONTROL_{name}', False)
            self.changeStyleDynamic(button, 'running', value)

        for name, button in self.powerBoot.items():
            value = self.app.power.data.get(f'POWER_ON_BOOT.POWER_PORT_{name}', False)
            button.setChecked(value)

        for name, button in self.current.items():
            value = self.app.power.data.get(f'POWER_CURRENT.POWER_CURRENT_{name}', 0)
            button.setText(f'{value:4.2f}')

        for name, button in self.dew.items():
            value = self.app.power.data.get(f'DEW_PWM.DEW_{name}', 0)
            button.setText(f'{value:3.0f}')

        for name, button in self.label.items():
            value = self.app.power.data.get(f'POWER_CONTROL_LABEL.POWER_LABEL_{name}',
                                            f'Power {name}')
            button.setText(value)

        value = self.app.power.data.get('POWER_CONSUMPTION.CONSUMPTION_AVG_AMPS', 0)
        self.ui.consumptionAvgAmps.setText('{0:4.2f}'.format(value))
        value = self.app.power.data.get('POWER_CONSUMPTION.CONSUMPTION_AMP_HOURS', 0)
        self.ui.consumptionAmpHours.setText('{0:4.2f}'.format(value))
        value = self.app.power.data.get('POWER_CONSUMPTION.CONSUMPTION_WATT_HOURS', 0)
        self.ui.consumptionWattHours.setText('{0:4.2f}'.format(value))

        value = self.app.power.data.get('POWER_SENSORS.SENSOR_VOLTAGE', 0)
        self.ui.sensorVoltage.setText('{0:4.1f}'.format(value))
        value = self.app.power.data.get('POWER_SENSORS.SENSOR_CURRENT', 0)
        self.ui.sensorCurrent.setText('{0:4.2f}'.format(value))
        value = self.app.power.data.get('POWER_SENSORS.SENSOR_POWER', 0)
        self.ui.sensorPower.setText('{0:4.2f}'.format(value))

        value = self.app.power.data.get('DEW_CURRENT.DEW_CURRENT_A', 0)
        self.ui.dewCurrentA.setText('{0:4.2f}'.format(value))
        value = self.app.power.data.get('DEW_CURRENT.DEW_CURRENT_B', 0)
        self.ui.dewCurrentB.setText('{0:4.2f}'.format(value))
        value = self.app.power.data.get('DEW_CURRENT.DEW_CURRENT_C', 0)
        self.ui.dewCurrentC.setText('{0:4.2f}'.format(value))

        value1 = self.app.power.data.get('AUTO_DEW.INDI_ENABLED', False)
        value2 = self.app.power.data.get('AUTO_DEW.DEW_A', False)
        value3 = self.app.power.data.get('AUTO_DEW.DEW_B', False)
        value4 = self.app.power.data.get('AUTO_DEW.DEW_C', False)
        value = value1 or value2 or value3 or value4
        self.changeStyleDynamic(self.ui.autoDew, 'running', value)

        if self.app.power.data.get('FIRMWARE_INFO.VERSION', '1.4') > '1.4':
            value = self.app.power.data.get('ADJUSTABLE_VOLTAGE.ADJUSTABLE_VOLTAGE_VALUE', 0)
            self.ui.adjustableOutput.setText(f'{value:4.1f}')

            for name, button in self.portUSB.items():
                value = self.app.power.data.get(f'USB_PORT_CONTROL.PORT_{name}', False)
                self.changeStyleDynamic(button, 'running', value)

        else:
            value = self.app.power.data.get('USB_HUB_CONTROL.INDI_ENABLED', False)
            self.changeStyleDynamic(self.ui.hubUSB, 'running', value)

        return True

    def setDew(self):
        """
        setDew send the new dew value to power. as self.sender() gives only the object
        back, which is in case of an QLineEdit a wrapper, we have to go to the parent
        object to get the line edit object directly

        :return: true fot test purpose
        """

        for name, button in self.dew.items():
            if button != self.sender().parent():
                continue

            actValue = valueToInt(button.text())

            if actValue is None:
                return False

            dlg = PyQt5.QtWidgets.QInputDialog()
            value, ok = dlg.getInt(self,
                                   f'Set dew PWM {name}',
                                   'Value (0-100):',
                                   actValue,
                                   0,
                                   100,
                                   10,
                                   )

            if not ok:
                return False

            suc = self.app.power.sendDew(port=name, value=value)
            return suc

    def togglePowerPort(self):
        """
        togglePowerPort  toggles the state of the power switch
        :return: success
        """

        for name, button in self.powerOnOFF.items():
            if button != self.sender():
                continue
            suc = self.app.power.togglePowerPort(port=name)
            return suc

    def togglePowerBootPort(self):
        """
        togglePowerPort  toggles the state of the power switch
        :return: success
        """

        for name, button in self.powerBoot.items():
            if button != self.sender():
                continue
            suc = self.app.power.togglePowerPortBoot(port=name)
            return suc

    def toggleHubUSB(self):
        """
        toggleHubUSB

        :return: true fot test purpose
        """
        suc = self.app.power.toggleHubUSB()
        return suc

    def togglePortUSB(self):
        """
        toggleUSBPort  toggles the state of the power switch
        :return: success
        """

        for name, button in self.portUSB.items():
            if button != self.sender():
                continue
            suc = self.app.power.togglePortUSB(port=name)
            return suc

    def toggleAutoDew(self):
        """
        toggleAutoDew

        :return: true fot test purpose
        """

        suc = self.app.power.toggleAutoDew()
        return suc

    def setAdjustableOutput(self):
        """
        setAdjustableOutput

        :return: true fot test purpose
        """

        actValue = float(self.ui.adjustableOutput.text())

        dlg = PyQt5.QtWidgets.QInputDialog()
        value, ok = dlg.getDouble(self,
                                  'Set Voltage Output',
                                  'Value (3-12):',
                                  actValue,
                                  3,
                                  12,
                                  1,
                                  PyQt5.QtCore.Qt.WindowFlags(),
                                  0.1,
                                  )

        if not ok:
            return False

        suc = self.app.power.sendAdjustableOutput(value=value)
        return suc

    def rebootUPB(self):
        """

        :return: true fot test purpose
        """

        suc = self.app.power.reboot()
        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import logging

# external packages
import numpy as np
from skyfield.api import Angle

# local imports
from mountcontrol.convert import valueToFloat


class Geometry(object):
    """
    The class Geometry contains all necessary geometric calculations and
    parameters for mount orientation, dome slit correction or 3D animation of
    telescope.

        >>> geometry = Geometry()

    transformations are defined as homogeneous matrices to be able to calculate
    translations and rotation in the same manner. therefore the system has 4
    dimensions

    for the transformation between homogeneous vectors (vh) and kartesian vectors
    (vk):
    vk = vh[0:2]/vh[3]
    vh = [vk, 1]

    """

    __all__ = ['Geometry',
               ]

    log = logging.getLogger(__name__)

    geometryData = {
        '10micron GM1000HPS': {
            'offBaseAltAxisX': -0.005,
            'offBaseAltAxisZ': 0.095,
            'offAltAxisGemX': 0.156,
            'offAltAxisGemZ': 0.12,
            'offGemPlate': 0.15,
        },
        '10micron GM2000HPS': {
            'offBaseAltAxisX': -0.102,
            'offBaseAltAxisZ': 0.132,
            'offAltAxisGemX': 0.304,
            'offAltAxisGemZ': 0.0,
            'offGemPlate': 0.175,
        },
        '10micron GM3000HPS': {
            'offBaseAltAxisX': -0.122,
            'offBaseAltAxisZ': 0.167,
            'offAltAxisGemX': 0.363,
            'offAltAxisGemZ': 0.0,
            'offGemPlate': 0.233,
        },
        '10micron GM4000HPS': {
            'offBaseAltAxisX': -0.128,
            'offBaseAltAxisZ': 0.19,
            'offAltAxisGemX': 0.413,
            'offAltAxisGemZ': 0.0,
            'offGemPlate': 0.300,
        },
    }

    def __init__(self, obsSite=None):

        self.obsSite = obsSite
        self.offBaseAltAxisX = 0
        self.offBaseAltAxisZ = 0
        self.offAltAxisGemX = 0
        self.offAltAxisGemZ = 0
        self.offGemPlate = 0
        self.altAdj = 0
        self.azAdj = 0

        self._offNorth = 0
        self._offEast = 0
        self._offVert = 0
        self._offVertGEM = 0
        self._domeRadius = 1
        self._offGEM = 0
        self._offLAT = 0
        self._offPlateOTA = 0

        self.transMatrix = None
        self.transVector = None

    @property
    def domeRadius(self):
        return self._domeRadius

    @domeRadius.setter
    def domeRadius(self, value):
        self._domeRadius = valueToFloat(value)

    @property
    def offNorth(self):
        return self._offNorth

    @offNorth.setter
    def offNorth(self, value):
        self._offNorth = valueToFloat(value)

    @property
    def offEast(self):
        return self._offEast

    @offEast.setter
    def offEast(self, value):
        self._offEast = valueToFloat(value)

    @property
    def offVert(self):
        return self._offVert

    @offVert.setter
    def offVert(self, value):
        self._offVert = valueToFloat(value)
        if self.obsSite.location is None:
            self.log.debug('offVert called without lat')
            return

        lat = self.obsSite.location.latitude.radians
        val = valueToFloat(value) + self.offBaseAltAxisZ
        val += np.sin(abs(lat)) * self.offAltAxisGemX
        self._offVertGEM = val

    @property
    def offVertGEM(self):
        return self._offVertGEM

    @offVertGEM.setter
    def offVertGEM(self, value):
        self._offVertGEM = valueToFloat(value)
        if self.obsSite.location is None:
            self.log.debug('offVertGEM called without lat')
            return

        lat = self.obsSite.location.latitude.radians
        val = valueToFloat(value) - self.offBaseAltAxisZ
        val -= np.sin(abs(lat)) * self.offAltAxisGemX
        self._offVert = val

    @property
    def offGEM(self):
        return self._offGEM

    @offGEM.setter
    def offGEM(self, value):
        self._offGEM = valueToFloat(value)
        self._offPlateOTA = self._offGEM - self.offGemPlate

    @property
    def offLAT(self):
        return self._offLAT

    @offLAT.setter
    def offLAT(self, value):
        self._offLAT = valueToFloat(value)

    @property
    def offPlateOTA(self):
        return self._offPlateOTA

    @offPlateOTA.setter
    def offPlateOTA(self, value):
        self._offPlateOTA = valueToFloat(value)
        self._offGEM = self._offPlateOTA + self.offGemPlate

    def initializeGeometry(self, mountType):
        """
        initializeGeometry takes the mount type as string and searches for the
        right parameters in his database. If found it populates the parameters
        for geometry calculation and returns True otherwise false

        :param mountType: string from mount
        :return: success
        """

        if mountType not in self.geometryData:
            self.log.error(f'[{mountType}] not in database')
            return False

        else:
            self.log.debug(f'using [{mountType}] geometry')

        self.offBaseAltAxisX = self.geometryData[mountType]['offBaseAltAxisX']
        self.offBaseAltAxisZ = self.geometryData[mountType]['offBaseAltAxisZ']
        self.offAltAxisGemX = self.geometryData[mountType]['offAltAxisGemX']
        self.offAltAxisGemZ = self.geometryData[mountType]['offAltAxisGemZ']
        self.offGemPlate = self.geometryData[mountType]['offGemPlate']

        return True

    @staticmethod
    def transformRotX(rotX, degrees=False):
        """
        :param rotX: rotation angle
        :param degrees:
        :return: homogeneous transformation matrix
        """
        if isinstance(rotX, Angle):
            rot = rotX.radians
        else:
            if degrees:
                rot = np.radians(rotX)
            else:
                rot = rotX

        tCos = np.cos(rot)
        tSin = np.sin(rot)

        T = np.array([[1, 0, 0, 0],
                      [0, tCos, -tSin, 0],
                      [0, tSin, tCos, 0],
                      [0, 0, 0, 1]])
        return T

    @staticmethod
    def transformRotY(rotY, degrees=False):
        """
        :param rotY: rotation angle
        :param degrees:
        :return: homogeneous transformation matrix
        """
        if isinstance(rotY, Angle):
            rot = rotY.radians
        else:
            if degrees:
                rot = np.radians(rotY)
            else:
                rot = rotY

        tCos = np.cos(rot)
        tSin = np.sin(rot)

        T = np.array([[tCos, 0, tSin, 0],
                      [0, 1, 0, 0],
                      [-tSin, 0, tCos, 0],
                      [0, 0, 0, 1]])
        return T

    @staticmethod
    def transformRotZ(rotZ, degrees=False):
        """
        :param rotZ: rotation angle
        :param degrees:
        :return: homogeneous transformation matrix
        """
        if isinstance(rotZ, Angle):
            rot = rotZ.radians
        else:
            if degrees:
                rot = np.radians(rotZ)
            else:
                rot = rotZ

        tCos = np.cos(rot)
        tSin = np.sin(rot)

        T = np.array([[tCos, -tSin, 0, 0],
                      [tSin, tCos, 0, 0],
                      [0, 0, 1, 0],
                      [0, 0, 0, 1]])
        return T

    @staticmethod
    def transformTranslate(vector):
        """
        :param vector: translation
        :return: homogeneous transformation matrix
        """
        T = np.array([[1, 0, 0, vector[0]],
                      [0, 1, 0, vector[1]],
                      [0, 0, 1, vector[2]],
                      [0, 0, 0, 1]])
        return T

    def calcTransformationMatrices(self, ha=None, dec=None, lat=None, pierside='W'):
        """
        :param ha: hour angle in radians
        :param dec: declination in radians
        :param lat: latitude of observation site in radians
        :param pierside: 'W' or 'E' for setting the right HA
        :return: altDome, azDome: the real pointing angles for OTA in dome
        """
        if not isinstance(ha, Angle):
            return None, None, None, None, None
        if not isinstance(dec, Angle):
            return None, None, None, None, None
        if not isinstance(lat, Angle):
            return None, None, None, None, None

        text = f'HA:{ha.hours}, DEC:{dec.degrees}, LAT:{lat.degrees}, '
        text += f'pierside:{pierside} ,'
        text += f'offGEM:{self.offGEM}, offPlateOTA:{self.offPlateOTA}, '
        text += f'offNorth:{self.offNorth}, offEast:{self.offEast}, '
        text += f'offVert:{self.offVert}, offLAT:{self.offLAT}, '
        text += f'domeRadius:{self.domeRadius}'
        self.log.trace(text)

        ha = ha.radians
        dec = dec.radians
        lat = lat.radians

        # the equator of the dome and it's middle point of the hemisphere is
        # zero point for coordinate system.
        P0 = np.array([0, 0, 0, 1])

        # next adjustment if the position of the base of mount in relation to
        # the offset in north is x and the offset in east is -y. in vertical
        # direction up is positive, which means the mount is above the middle
        # point of dome
        vec0 = [self.offNorth,
                -self.offEast,
                self.offVert]
        T0 = self.transformTranslate(vec0)
        P1 = np.dot(T0, P0)

        # the rotation around z axis (polar direction) to adjust the orientation
        # of the mount to true north, if the fixed pier does not make it.
        # turning counterclockwise is positive
        T1 = np.dot(T0, self.transformRotZ(self.azAdj))
        P2 = np.dot(T1, P0)

        # next we have the translation between the base plate and the rotation
        # axis for the altitude adjustment. the axis is normally above the plate (
        # positive z) and sometime shifted. shift north mean positive x
        vec2 = [self.offBaseAltAxisX,
                0,
                self.offBaseAltAxisZ]
        T2 = np.dot(T1, self.transformTranslate(vec2))
        P3 = np.dot(T2, P0)

        # next step is the rotation around y axis for compensation of latitude.
        # the adjustment angle compensates the latitude. for lat = 0 we have 0
        # degree correction an below north pole there is 90 deg correction needed.
        # so phi = -lat because it's a turn counterclockwise around Y, all angles
        # are in radians
        self.altAdj = -abs(lat)
        T3 = np.dot(T2, self.transformRotY(self.altAdj))
        P4 = np.dot(T3, P0)

        # next is the translation fom the rotation axis of the lat compensation of
        # the mount to the GEM point of the mount. GEM means the crossing of the
        # ra and dec axis of the german equatorial mount basically there should be
        # a translation in x/z plane up and forward (to north)
        vec4 = [self.offAltAxisGemX,
                0,
                self.offAltAxisGemZ]
        T4 = np.dot(T3, self.transformTranslate(vec4))
        P5 = np.dot(T4, P0)

        # next is the rotation around the ra axis of the mount, this is rotation
        # around x this should be (as we don't track) measured in HA, where HA = 6
        # / 18 h is North depending of the pierside the direction is clockwise,
        # turning to west over time
        if pierside == 'E':
            value = - ha + np.radians(6 / 24 * 360)

        else:
            value = - ha + np.radians(18 / 24 * 360)

        T5 = np.dot(T4, self.transformRotX(value))
        P6 = np.dot(T5, P0)

        # the rotation around dec axis of the mount is next step. this rotation is
        # around z axis. dec = 0 means rectangular directing scope. direction
        # changes due to position of the mount and pierside
        value = dec - np.radians(90)
        if pierside == 'E':
            value = -value

        T6 = np.dot(T5, self.transformRotZ(value))
        P7 = np.dot(T6, P0)

        # the translation from GEM to the center of the ota, this should be a
        # translation in z. it consists of two parts: the user OTA measures and the
        # distance from GEM to base plate, which is depending on mount type and is
        # fixed and secondly the distance between the mount plate and the OTA line
        # of sight axis
        vec6 = [0,
                0,
                self.offPlateOTA + self.offGemPlate]
        T7 = np.dot(T6, self.transformTranslate(vec6))
        P8 = np.dot(T7, P0)

        # the translation from center of the ota, to the side if a second ota is
        # installed but not centered should be a translation in y. if the ota is
        # de-centered looking in the direction of the ota (out of the hemisphere)
        # to the right it's a negative y otherwise a positive one

        vec7 = [0,
                - self.offLAT,
                0]
        T8 = np.dot(T7, self.transformTranslate(vec7))
        P9 = np.dot(T8, P0)

        # calculating the direction of OTA for dome geometry. pointing direction is
        # in x axis, length is standard 1. P10 would be the head of the vector,
        # P9 is the base of the vector. subtracting gives the resulting direction
        # of the vector which is in PD
        vec9 = [1, 0, 0]
        T9 = np.dot(T8, self.transformTranslate(vec9))
        P10 = np.dot(T9, P0)
        PD = (P10 - P9)[:-1]

        # calculating the crossing point between view of sight of the OTA and the
        # dome hemisphere using the traditional p-q formula
        # if you have g: v = v0 + t * tDir and hem: x^2 + y^2 + z^2 = r^2
        # with v0 = PB and tDir = PD you get t1 and t2 from p-q formula as
        # t1 = -p/2 + sqr (p^2/4 - q), t2 = -p/2 - sqr (p^2/4 - q)
        # for the right point whe have to choose the one, which crosses the
        # hemisphere above the OTA point as we normally are looking upward.
        # and we have to use scalar product on a cartesian system
        PB = P9[:-1]

        p = 2 * np.dot(PD, PB)
        q = (np.dot(PB, PB) - self.domeRadius**2)

        self.log.trace(f'Geometry calc p:[{p}], q:[{q}]')

        # there should be always a reasonable solution
        if (p * p / 4 - q) < 0:
            self.log.error('Geometry solution impossible')
            return None, None, None, None, None

        t1 = - p / 2 + np.sqrt(p * p / 4 - q)
        t2 = - p / 2 - np.sqrt(p * p / 4 - q)

        # we choose the positive solution as we look in the positive direction and
        # can omit the view to the back

        self.log.debug(f'Solutions: [{t1}], [{t2}]')

        intersect = PB + np.dot(t1, PD)

        # simplify the names and calculate the geometry angles based on the
        # coordinates of the intersection between line of sight and dome
        # hemisphere. as we have y to the left because of the right turning
        # homogeneous coordinate system, we have to negate the angle of azimuth.
        # we use arctan2 for getting the full quadrants and shifting
        # the angle in an value range from 0 .. 2pi

        x = intersect[0]
        y = intersect[1]
        z = intersect[2]

        value = np.mod(- np.arctan2(y, x), 2 * np.pi)
        azDome = Angle(radians=value)
        base = np.sqrt(x * x + y * y)

        altDome = Angle(radians=np.arctan2(z, base))

        self.transMatrix = [T0, T1, T2, T3, T4, T5, T6, T7, T8, T9]
        self.transVector = [P0[:-1], P1[:-1], P2[:-1], P3[:-1], P4[:-1],
                            P5[:-1], P6[:-1], P7[:-1], P8[:-1], P9[:-1],
                            P10[:-1]]

        self.log.trace(f'az:{azDome}, alt:{altDome}')

        return altDome, azDome, x, y, z



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages

# local import


class SettParkPos(object):
    """
    """

    def __init__(self, app=None, ui=None, clickable=None):
        if app:
            self.app = app
            self.ui = ui
            self.clickable = clickable

        self.posButtons = list()
        self.posTexts = list()
        self.posAlt = list()
        self.posAz = list()
        self.posSaveButtons = list()

        self.setupParkPosGui()
        for posText in self.posTexts:
            posText.editingFinished.connect(self.updateParkPosButtonText)
        for button in self.posButtons:
            button.clicked.connect(self.slewToParkPos)
        for button in self.posSaveButtons:
            button.clicked.connect(self.saveActualPosition)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """

        config = self.app.config['mainW']

        for i, textField in enumerate(self.posTexts):
            keyConfig = 'posText{0:1d}'.format(i)
            textField.setText(config.get(keyConfig, 'Park Pos {0:1d}'.format(i)))
        for i, textField in enumerate(self.posAlt):
            keyConfig = 'posAlt{0:1d}'.format(i)
            textField.setText(config.get(keyConfig, '-'))
        for i, textField in enumerate(self.posAz):
            keyConfig = 'posAz{0:1d}'.format(i)
            textField.setText(config.get(keyConfig, '-'))
        self.updateParkPosButtonText()
        self.ui.parkMountAfterSlew.setChecked(config.get('parkMountAfterSlew', False))

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """

        config = self.app.config['mainW']

        for i, textField in enumerate(self.posTexts):
            keyConfig = 'posText{0:1d}'.format(i)
            config[keyConfig] = textField.text()
        for i, textField in enumerate(self.posAlt):
            keyConfig = 'posAlt{0:1d}'.format(i)
            config[keyConfig] = textField.text()
        for i, textField in enumerate(self.posAz):
            keyConfig = 'posAz{0:1d}'.format(i)
            config[keyConfig] = textField.text()
        config['parkMountAfterSlew'] = self.ui.parkMountAfterSlew.isChecked()

        return True

    def setupParkPosGui(self):
        """
        setupRelayGui handles the modeldata of list for relay handling. to keep many relay in
        order i collect them in the list for list handling afterwards.

        :return: True for test purpose
        """
        # generate the button list and text entry for later use
        for i in range(0, 10):
            self.posButtons.append(eval('self.ui.posButton{0:1d}'.format(i)))
            self.posTexts.append(eval('self.ui.posText{0:1d}'.format(i)))
            self.posAlt.append(eval('self.ui.posAlt{0:1d}'.format(i)))
            self.posAz.append(eval('self.ui.posAz{0:1d}'.format(i)))
            self.posSaveButtons.append(eval('self.ui.posSave{0:1d}'.format(i)))

        return True

    def updateParkPosButtonText(self):
        """
        updateParkPosButtonText updates the text in the gui button if we change texts in
        the gui.

        :return: true for test purpose
        """

        for button, textField in zip(self.posButtons, self.posTexts):
            button.setText(textField.text())

        return True

    def parkAtPos(self):
        """
        :return:
        """
        self.app.mount.signals.slewFinished.disconnect(self.parkAtPos)
        suc = self.app.mount.obsSite.parkOnActualPosition()

        if not suc:
            self.app.message.emit('Cannot park at current position', 2)

        return suc

    def slewToParkPos(self):
        """
        slewToParkPos takes the configured data from park positions menu and slews the mount
        to the targeted alt az coordinates and stops tracking. actually there is no chance to
        park the mount directly.

        :return: success
        """

        for button, posText, alt, az in zip(self.posButtons,
                                            self.posTexts,
                                            self.posAlt,
                                            self.posAz,
                                            ):

            if button != self.sender():
                continue

            try:
                altValue = float(alt.text())
                azValue = float(az.text())

            except Exception as e:
                self.log.critical(f'no usable values in data: error {e}')
                self.app.message.emit('Missing correct entries in settings', 2)
                return False

            posTextValue = posText.text()

            if altValue < -5:
                altValue = -5
            elif altValue > 90:
                altValue = 90
            azValue = azValue % 360

            suc = self.app.mount.obsSite.setTargetAltAz(alt_degrees=altValue,
                                                        az_degrees=azValue)

            if suc:
                suc = self.app.mount.obsSite.startSlewing(slewType='notrack')

            if not suc:
                self.app.message.emit(f'Cannot slew to [{posTextValue}]', 2)
                return False

            self.app.message.emit(f'Slew to [{posTextValue}]', 0)

            if not self.ui.parkMountAfterSlew.isChecked():
                return True

            self.app.mount.signals.slewFinished.connect(self.parkAtPos)

            return True

        return False

    def saveActualPosition(self):
        """
        saveActualPosition takes the actual mount position alt/az and stores it in the alt az
        fields in the gui for persistence.

        :return: success
        """

        obs = self.app.mount.obsSite

        if not obs.Alt:
            return False
        if not obs.Az:
            return False

        for button, alt, az in zip(self.posSaveButtons,
                                   self.posAlt,
                                   self.posAz,
                                   ):

            if button != self.sender():
                continue

            alt.setText(f'{obs.Alt.degrees:3.0f}')
            az.setText(f'{obs.Az.degrees:3.0f}')

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages
from PyQt5.QtGui import QVector3D
from PyQt5.Qt3DCore import QEntity

# local import
from gui.extWindows.simulator.materials import Materials
from gui.extWindows.simulator import tools


class SimulatorDome:

    __all__ = ['SimulatorDome',
               ]

    def __init__(self, app):
        super().__init__()

        self.app = app
        self.model = {}
        self.modelRoot = None

    def create(self, rEntity, show):
        """
        :param rEntity:
        :param show:
        :return:
        """
        if self.model:
            self.modelRoot.setParent(None)

        self.model.clear()
        if not show:
            return False

        self.modelRoot = QEntity(rEntity)
        self.model = {
            'domeFloor': {
                'parent': 'ref',
                'source': 'dome-floor.stl',
                'scale': [1, 1, 1],
                'mat': Materials().aluminiumGrey,
            },
            'domeWall': {
                'parent': 'ref',
                'source': 'dome-wall.stl',
                'scale': [1, 1, 1],
                'mat': Materials().dome3,
            },
            'domeSphere': {
                'parent': 'ref',
                'source': 'dome-sphere.stl',
                'scale': [1, 1, 1],
                'mat': Materials().dome1,
            },
            'domeSlit1': {
                'parent': 'domeSphere',
                'source': 'dome-slit1.stl',
                'scale': [1, 1, 1],
                'mat': Materials().dome2,
            },
            'domeSlit2': {
                'parent': 'domeSphere',
                'source': 'dome-slit2.stl',
                'scale': [1, 1, 1],
                'mat': Materials().dome2,
            },
            'domeDoor1': {
                'parent': 'domeSphere',
                'source': 'dome-door1.stl',
                'scale': [1, 1, 1],
                'mat': Materials().dome2,
            },
            'domeDoor2': {
                'parent': 'domeSphere',
                'source': 'dome-door2.stl',
                'scale': [1, 1, 1],
                'mat': Materials().dome2,
            },
        }
        for name in self.model:
            tools.linkModel(self.model, name, self.modelRoot)
        return True

    def setTransparency(self, transparent):
        """
        :return: True for test purpose
        """
        if not self.model:
            return False

        entities = {
            'domeWall': {
                'trans': Materials().dome3t,
                'solid': Materials().dome3
            },
            'domeSphere': {
                'trans': Materials().dome1t,
                'solid': Materials().dome1
            },
            'domeSlit1': {
                'trans': Materials().dome2t,
                'solid': Materials().dome2
            },
            'domeSlit2': {
                'trans': Materials().dome2t,
                'solid': Materials().dome2
            },
            'domeDoor1': {
                'trans': Materials().dome2t,
                'solid': Materials().dome2
            },
            'domeDoor2': {
                'trans': Materials().dome2t,
                'solid': Materials().dome2
            },
        }
        for entity in entities:
            if transparent:
                self.model[entity]['e'].addComponent(entities[entity]['trans'])

            else:
                self.model[entity]['e'].addComponent(entities[entity]['solid'])
        return True

    def updateSettings(self):
        """
        updateSettings resize parts depending on the setting made in the dome tab. likewise
        some transformations have to be reverted as they are propagated through entity linking.

        :return:
        """
        if not self.model:
            return False

        radius = self.app.mount.geometry.domeRadius * 1000
        scale = 1 + (radius - 1250) / 1250
        corrZ = - (scale - 1) * 800
        self.model['domeFloor']['t'].setScale3D(QVector3D(scale, scale, 1))
        self.model['domeWall']['t'].setScale3D(QVector3D(scale, scale, 1))
        self.model['domeSphere']['t'].setScale3D(QVector3D(scale, scale, scale))
        self.model['domeSphere']['t'].setTranslation(QVector3D(0, 0, corrZ))
        return True

    def updatePositions(self):
        """
        updateDome moves dome components
        you normally have to revert your transformation in linked entities if
        they have fixed sizes because they propagate transformations.
        for the shutter i would like to keep the width setting unscaled with
        increasing dome radius

        :return: success
        """
        if not self.model:
            return False

        if 'ABS_DOME_POSITION.DOME_ABSOLUTE_POSITION' in self.app.dome.data:
            az = self.app.dome.data['ABS_DOME_POSITION.DOME_ABSOLUTE_POSITION']
            self.model['domeSphere']['t'].setRotationZ(-az)

        if 'DOME_SHUTTER.SHUTTER_OPEN' in self.app.dome.data:
            isOpen = self.app.dome.data['DOME_SHUTTER.SHUTTER_OPEN']
            radius = self.app.mount.geometry.domeRadius * 1000
            scale = 1 + (radius - 1250) / 1250
            width = self.app.dome.domeShutterWidth * 1000
            scaleSlit = (1 + (width - 600) / 600 / 2) * 0.9
            shiftShutter = width / 2 / scale

            self.model['domeSlit1']['t'].setScale3D(QVector3D(1, scaleSlit, 1))
            self.model['domeSlit2']['t'].setScale3D(QVector3D(1, scaleSlit, 1))

            if isOpen:
                self.model['domeDoor1']['t'].setTranslation(QVector3D(0,
                                                                      shiftShutter,
                                                                      0))
                self.model['domeDoor2']['t'].setTranslation(QVector3D(0,
                                                                      -shiftShutter,
                                                                      0))
            else:
                self.model['domeDoor1']['t'].setTranslation(QVector3D(0, 0, 0))
                self.model['domeDoor2']['t'].setTranslation(QVector3D(0, 0, 0))

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python3
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
from setuptools import setup
from pathlib import Path
import platform

setup(
    name='mountwizzard4',
    version='1.0.1',
    packages=[
        'mw4',
        'mw4.base',
        'mw4.gui',
        'mw4.indibase',
        'mw4.gui.extWindows',
        'mw4.gui.extWindows.simulator',
        'mw4.gui.mainWmixin',
        'mw4.gui.mainWindow',
        'mw4.gui.utilities',
        'mw4.gui.widgets',
        'mw4.logic.astrometry',
        'mw4.logic.automation',
        'mw4.logic.cover',
        'mw4.logic.databaseProcessing',
        'mw4.logic.dome',
        'mw4.logic.environment',
        'mw4.logic.imaging',
        'mw4.logic.measure',
        'mw4.logic.modeldata',
        'mw4.logic.powerswitch',
        'mw4.logic.remote',
        'mw4.logic.telescope',
        'mw4.mountcontrol',
        'mw4.resource',
    ],
    python_requires='>=3.7.0, <3.10',
    install_requires=[
        'numpy==1.19.3',
        'pillow==8.1.0',
        'matplotlib==3.3.3',
        'pyerfa==1.7.1.1',
        'astropy==4.2',
        'scipy==1.5.4',
        'sep==1.1.1',
        'requests==2.24.0',
        'requests_toolbelt==0.9.1',
        'skyfield==1.36',
        'sgp4>=2.1',
        'qimage2ndarray==1.8.3',
        'importlib_metadata==3.4.0',
        'deepdiff==5.0.2',
        'colour_demosaicing==0.1.5',
        'wakeonlan==1.1.6',
        'pybase64==1.1.1'
    ]
    + (['pywin32==300'] if "Windows" == platform.system() else [])
    + (['pywinauto==0.6.8'] if "Windows" == platform.system() else [])
    + (['PyQt5==5.15.2'] if platform.machine() not in ['armv7l'] else [])
    + (['PyQt3D==5.15.2'] if platform.machine() not in ['armv7l', 'aarch64'] else [])
    + (['PyQtWebEngine==5.15.2'] if platform.machine() not in ['armv7l', 'aarch64'] else []),

    url='https://github.com/mworion/MountWizzard4',
    license='APL 2.0',
    author='mworion',
    author_email='michael@wuertenberger.org',
    description='Tool for managing 10micron mounts',
    long_description=Path("README.rst").read_text(encoding="utf-8"),
    long_description_content_type="text/x-rst",
    project_urls={
        'Documentation': 'https://mountwizzard4.readthedocs.io',
        'Source Code': 'https://github.com/mworion/mountwizzard4',
        'Bug Tracker': 'https://github.com/mworion/mountwizzard4/issues',
        'Discussions': 'https://github.com/mworion/MountWizzard4/discussions',
        'Channel': 'https://www.youtube.com/user/orion49m/featured',
        'Forum': 'https://www.10micron.eu/forum/',
    },
    zip_safe=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: MacOS X',
        'Environment :: Other Environment',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: POSIX :: Linux',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Utilities',
        'Topic :: Scientific/Engineering :: Astronomy',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Natural Language :: German',
        'Topic :: Documentation :: Sphinx',
    ]
)



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import logging

# external packages

# local imports
from mountcontrol.connection import Connection
from mountcontrol.convert import valueToInt


class Firmware(object):
    """
    The class Firmware inherits all information and handling of firmware
    attributes of the connected mount and provides the abstracted interface
    to a 10 micron mount.

        >>> fw = Firmware(host='')
    """

    __all__ = ['Firmware',
               ]

    log = logging.getLogger(__name__)

    def __init__(self,
                 host=None,
                 ):

        self.host = host
        self._product = None
        self._vString = None
        self._hardware = None
        self._date = None
        self._time = None

    @property
    def product(self):
        return self._product

    @product.setter
    def product(self, value):
        self._product = value

    @property
    def vString(self):
        return self._vString

    @vString.setter
    def vString(self, value):
        if not isinstance(value, str):
            self._vString = None
            return
        if not value.count('.') > 0:
            self._vString = None
            return
        if not all(valueToInt(x) for x in value.split('.')):
            self._vString = None
            return
        self._vString = value

    @property
    def hardware(self):
        return self._hardware

    @hardware.setter
    def hardware(self, value):
        self._hardware = value

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, value):
        self._date = value

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, value):
        self._time = value

    def number(self):
        if not self._vString:
            return None

        parts = self._vString.split('.')
        try:
            if len(parts) == 3:
                value = int(parts[0]) * 10000 + int(parts[1]) * 100 + int(parts[2])

            elif len(parts) == 2:
                value = int(parts[0]) * 10000 + int(parts[1]) * 100

            else:
                value = None

        except Exception as e:
            self.log.warning(f'error: {e}, malformed value: {parts}')
            return None

        else:
            return value

    def checkNewer(self, compare):
        """
        Checks if the provided FW number is newer than the one of the mount

        :param compare:     fw number to test as int
        :return:            True if newer / False
        """

        value = self.number()
        if value:
            return compare > value
        else:
            return None

    def parse(self, response, numberOfChunks):
        """
        Parsing the polling slow command.

        :param response:        data load from mount
        :param numberOfChunks:
        :return: success:       True if ok, False if not
        """

        if len(response) != numberOfChunks:
            self.log.warning('wrong number of chunks')
            return False
        self.date = response[0]
        self.vString = response[1]
        self.product = response[2]
        self.time = response[3]
        self.hardware = response[4]
        return True

    def poll(self):
        """
        Sending the polling slow command. As the mount need polling the data,
        I send a set of commands to get the data back to be able to process
        and store it.

        :return: success:   True if ok, False if not
        """

        conn = Connection(self.host)
        commandString = ':U2#:GVD#:GVN#:GVP#:GVT#:GVZ#'
        suc, response, chunks = conn.communicate(commandString)
        if not suc:
            return False
        suc = self.parse(response, chunks)
        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages
from PyQt5.QtGui import QVector3D
from PyQt5.Qt3DCore import QEntity
from PyQt5.Qt3DExtras import QSphereMesh

# local import
from gui.extWindows.simulator.materials import Materials
from gui.extWindows.simulator import tools


class SimulatorPointer:

    __all__ = ['SimulatorPointer',
               ]

    def __init__(self, app):
        super().__init__()

        self.app = app
        self.model = {}
        self.modelRoot = None

    def create(self, rEntity, show):
        """
        dict {'name of model': {'parent': }}

        :param rEntity: root of the 3D models
        :param show: root of the 3D models
        :return: success
        """

        if self.model:
            self.modelRoot.setParent(None)

        self.model.clear()

        if not show:
            return False

        self.modelRoot = QEntity(rEntity)

        self.model = {
            'pointer': {
                'parent': None,
                'source': [QSphereMesh(), 50, 30, 30],
                'scale': [1, 1, 1],
                'mat': Materials().pointer,
            },
        }

        for name in self.model:
            tools.linkModel(self.model, name, self.modelRoot)

        return True

    def updatePositions(self):
        """
        updatePositions calculates the crossing point of the actual telescope view and the
        dome if present.

        :return:
        """

        if not self.model:
            return False

        if not self.app.mount.obsSite.haJNow:
            return False

        lat = self.app.mount.obsSite.location.latitude
        ha = self.app.mount.obsSite.haJNow
        dec = self.app.mount.obsSite.decJNow
        pierside = self.app.mount.obsSite.pierside

        geometry = self.app.mount.geometry
        _, _, x, y, z = geometry.calcTransformationMatrices(ha=ha,
                                                            dec=dec,
                                                            lat=lat,
                                                            pierside=pierside)
        x = x * 1000
        y = y * 1000
        z = z * 1000 + 1000

        self.model['pointer']['t'].setTranslation(QVector3D(x, y, z))

        return True
